"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

This module implements the BillModel, which represents an Invoice received from a Supplier/Vendor, on which
the Vendor states the amount owed by the recipient for the purposes of supplying goods and/or services.
In addition to tracking the bill amount, it tracks the paid and due amount.

"""
import warnings
from datetime import date, datetime
from decimal import Decimal
from typing import Union, Optional, Tuple, Dict, List
from uuid import uuid4, UUID

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import models, transaction, IntegrityError
from django.db.models import Q, Sum, F, Count, QuerySet, Manager
from django.db.models.signals import pre_save
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_ledger.io import ASSET_CA_CASH, ASSET_CA_PREPAID, LIABILITY_CL_ACC_PAYABLE
from django_ledger.io.io_core import get_localtime, get_localdate
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.entity import EntityModel
from django_ledger.models.items import ItemTransactionModelQuerySet, ItemTransactionModel, ItemModel, ItemModelQuerySet
from django_ledger.models.mixins import (
    CreateUpdateMixIn,
    AccrualMixIn,
    MarkdownNotesMixIn,
    PaymentTermsMixIn,
    ItemizeMixIn
)
from django_ledger.models.signals import (
    bill_status_draft,
    bill_status_in_review,
    bill_status_approved,
    bill_status_paid,
    bill_status_canceled,
    bill_status_void,
)
from django_ledger.models.utils import lazy_loader
from django_ledger.settings import (DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING, DJANGO_LEDGER_BILL_NUMBER_PREFIX,
                                    DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR)

UserModel = get_user_model()


class BillModelValidationError(ValidationError):
    pass


class BillModelQuerySet(QuerySet):
    """
    A custom defined QuerySet for the BillModel. This implements multiple methods or queries needed to get a filtered
    QuerySet based on the BillModel status. For example, we might want to have list of bills which are paid, unpaid,
    due ,overdue, approved or in draft stage. All these separate functions will assist in making such queries and
    building customized reports.
    """

    def for_user(self, user_model) -> 'BillModelQuerySet':
        """
        Fetches a QuerySet of BillModels that the UserModel as access to.
        May include BillModels from multiple Entities.

        The user has access to bills if:
            1. Is listed as Manager of Entity.
            2. Is the Admin of the Entity.

        Parameters
        __________
        user_model
            Logged in and authenticated django UserModel instance.

        Returns
        _______
        BillModelQuerySet
            Returns a BillModelQuerySet with applied filters.
        """
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(ledger__entity__admin=user_model) |
            Q(ledger__entity__managers__in=[user_model])
        )

    def draft(self):
        """
        Default status of any bill that is created.
        Draft bills do not impact the Ledger.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of draft bills only.
        """
        return self.filter(bill_status__exact=BillModel.BILL_STATUS_DRAFT)

    def in_review(self):
        """
        In review bills are those that need additional review or approvals before being approved.
        In review bills do not impact the Ledger.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of bills in review only.
        """
        return self.filter(bill_status__exact=BillModel.BILL_STATUS_REVIEW)

    def approved(self):
        """
        Approved bills are those that have been reviewed and are expected to be paid before the due date.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of approved bills only.
        """
        return self.filter(bill_status__exact=BillModel.BILL_STATUS_APPROVED)

    def paid(self):
        """
        Paid bills are those that have received 100% of the amount due.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of paid bills only.
        """
        return self.filter(bill_status__exact=BillModel.BILL_STATUS_PAID)

    def void(self):
        """
        Void bills are those that where rolled back after being approved.
        Void bills rollback all transactions by creating a new set of transactions posted on the date_void.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of void bills only.
        """
        return self.filter(bill_status__exact=BillModel.BILL_STATUS_VOID)

    def canceled(self):
        """
        Canceled bills are those that are discarded during the draft or in review status.
        These bills never had an impact on the books.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of canceled bills only.
        """
        return self.filter(bill_status__exact=BillModel.BILL_STATUS_CANCELED)

    def active(self):
        """
        Active bills are those that are approved or paid, which have
        impacted or have the potential to impact the Entity's Ledgers.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of active bills only.
        """
        return self.filter(
            Q(bill_status__exact=BillModel.BILL_STATUS_APPROVED) |
            Q(bill_status__exact=BillModel.BILL_STATUS_PAID)
        )

    def overdue(self):
        """
        Overdue bills are those which due date is in the past.

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of overdue bills only.
        """
        return self.filter(date_due__lt=get_localdate())

    def unpaid(self) -> 'BillModelQuerySet':
        """
        Unpaid bills are those that are approved but have not received 100% of the amount due.
        Equivalent to approved().

        Returns
        -------
        BillModelQuerySet
            Returns a QuerySet of paid bills only.
        """
        return self.filter(bill_status__exact=BillModel.BILL_STATUS_APPROVED)


class BillModelManager(Manager):
    """
    A custom-defined BillModelManager that will act as an interface to handling the initial DB queries
    to the BillModel. The default "get_queryset" has been overridden to refer the custom defined
    "BillModelQuerySet".
    """

    def get_queryset(self) -> BillModelQuerySet:
        qs = BillModelQuerySet(self.model, using=self._db)
        return qs.select_related(
            'ledger',
            'ledger__entity'
        )

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: EntityModel | str | UUID = None, **kwargs) -> BillModelQuerySet:
        """
        Fetches a QuerySet of BillModels associated with a specific EntityModel & UserModel.
        May pass an instance of EntityModel or a String representing the EntityModel slug.

        Parameters
        __________
        entity_model: str or EntityModel
            The entity slug or EntityModel used for filtering the QuerySet.

        Returns
        _______
        BillModelQuerySet
            Returns a BillModelQuerySet with applied filters.
        """
        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(ledger__entity=entity_model)
        elif isinstance(entity_model, str):
            qs =  qs.filter(ledger__entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(ledger__entity_id=entity_model)
        else:
            raise BillModelValidationError(
                'Must pass EntityModel, slug or UUID'
            )
        return qs


class BillModelAbstract(
    AccrualMixIn,
    ItemizeMixIn,
    PaymentTermsMixIn,
    MarkdownNotesMixIn,
    CreateUpdateMixIn
):
    """
    This is the main abstract class which the BillModel database will inherit from.
    The BillModel inherits functionality from the following MixIns:

        1. :func:`LedgerWrapperMixIn <django_ledger.models.mixins.LedgerWrapperMixIn>`
        2. :func:`PaymentTermsMixIn <django_ledger.models.mixins.PaymentTermsMixIn>`
        3. :func:`MarkdownNotesMixIn <django_ledger.models.mixins.MarkdownNotesMixIn>`
        4. :func:`CreateUpdateMixIn <django_ledger.models.mixins.CreateUpdateMixIn>`

    Attributes
    ----------
    uuid : UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().

    bill_number: str
        Auto assigned number at creation by generate_bill_number() function.
        Prefix be customized with DJANGO_LEDGER_BILL_NUMBER_PREFIX setting.
        Includes a reference to the Fiscal Year and a sequence number. Max Length is 20.

    bill_status: str
        Current status of the BillModel. Must be one of the choices as mentioned under "BILL_STATUS".
        By default , the status will be "Draft". Options are: Draft, In Review, Approved, Paid, Void or Canceled.

    xref: str
        This is the filed for capturing of any External reference number like the PO number of the buyer.
        Any other reference number like the Vendor code in buyer books may also be captured.

    vendor: :obj:`VendorModel`
        This is the foreign key reference to the VendorModel from whom the purchase has been made.

    additional_info: dict
        Any additional metadata about the BillModel may be stored here as a dictionary object.
        The data is serialized and stored as a JSON document in the Database.

    bill_items:
        A foreign key reference to the list of ItemTransactionModel that make the bill amount.

    ce_model: EstimateModel
        A foreign key to the BillModel associated EstimateModel for overall Job/Contract tracking.

    date_draft: date
        The draft date represents the date when the BillModel was first created. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.

    date_in_review: date
        The in review date represents the date when the BillModel was marked as In Review status.
        Will be null if BillModel is canceled during draft status. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.

    date_approved: date
        The approved date represents the date when the BillModel was approved. Will be null if BillModel is canceled.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.

    date_paid: date
        The paid date represents the date when the BillModel was paid and amount_due equals amount_paid.
        Will be null if BillModel is canceled. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.

    date_void: date
        The void date represents the date when the BillModel was void, if applicable.
        Will be null unless BillModel is void. Defaults to :func:`localdate <django.utils.timezone.localdate>`.

    date_canceled: date
        The canceled date represents the date when the BillModel was canceled, if applicable.
        Will be null unless BillModel is canceled. Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    """
    REL_NAME_PREFIX = 'bill'
    IS_DEBIT_BALANCE = False
    ALLOW_MIGRATE = True

    BILL_STATUS_DRAFT = 'draft'
    BILL_STATUS_REVIEW = 'in_review'
    BILL_STATUS_APPROVED = 'approved'
    BILL_STATUS_PAID = 'paid'
    BILL_STATUS_VOID = 'void'
    BILL_STATUS_CANCELED = 'canceled'

    BILL_STATUS = [
        (BILL_STATUS_DRAFT, _('Draft')),
        (BILL_STATUS_REVIEW, _('In Review')),
        (BILL_STATUS_APPROVED, _('Approved')),
        (BILL_STATUS_PAID, _('Paid')),
        (BILL_STATUS_CANCELED, _('Canceled')),
        (BILL_STATUS_VOID, _('Void'))
    ]
    """
    The different bill status options and their representation in the Database.
    """

    # todo: implement Void Bill (& Invoice)....
    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    entity_model = models.ForeignKey(
        'django_ledger.EntityModel',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        editable=False
    )
    bill_number = models.SlugField(max_length=20, verbose_name=_('Bill Number'), editable=False)
    bill_status = models.CharField(max_length=10,
                                   choices=BILL_STATUS,
                                   default=BILL_STATUS[0][0],
                                   verbose_name=_('Bill Status'))
    xref = models.SlugField(null=True, blank=True, verbose_name=_('External Reference Number'))
    vendor = models.ForeignKey('django_ledger.VendorModel',
                               on_delete=models.CASCADE,
                               verbose_name=_('Vendor'))

    cash_account = models.ForeignKey('django_ledger.AccountModel',
                                     on_delete=models.RESTRICT,
                                     null=True,
                                     blank=True,
                                     verbose_name=_('Cash Account'),
                                     related_name=f'{REL_NAME_PREFIX}_cash_account')
    prepaid_account = models.ForeignKey('django_ledger.AccountModel',
                                        on_delete=models.RESTRICT,
                                        null=True,
                                        blank=True,
                                        verbose_name=_('Prepaid Account'),
                                        related_name=f'{REL_NAME_PREFIX}_prepaid_account')
    unearned_account = models.ForeignKey('django_ledger.AccountModel',
                                         on_delete=models.RESTRICT,
                                         null=True,
                                         blank=True,
                                         verbose_name=_('Unearned Account'),
                                         related_name=f'{REL_NAME_PREFIX}_unearned_account')

    additional_info = models.JSONField(blank=True,
                                       null=True,
                                       default=dict,
                                       verbose_name=_('Bill Additional Info'))
    bill_items = models.ManyToManyField('django_ledger.ItemModel',
                                        through='django_ledger.ItemTransactionModel',
                                        through_fields=('bill_model', 'item_model'),
                                        verbose_name=_('Bill Items'))

    ce_model = models.ForeignKey('django_ledger.EstimateModel',
                                 on_delete=models.RESTRICT,
                                 null=True,
                                 blank=True,
                                 verbose_name=_('Associated Customer Job/Estimate'))

    date_draft = models.DateField(null=True, blank=True, verbose_name=_('Draft Date'))
    date_in_review = models.DateField(null=True, blank=True, verbose_name=_('In Review Date'))
    date_approved = models.DateField(null=True, blank=True, verbose_name=_('Approved Date'))
    date_paid = models.DateField(null=True, blank=True, verbose_name=_('Paid Date'))
    date_void = models.DateField(null=True, blank=True, verbose_name=_('Void Date'))
    date_canceled = models.DateField(null=True, blank=True, verbose_name=_('Canceled Date'))

    objects = BillModelManager.from_queryset(queryset_class=BillModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-updated']
        verbose_name = _('Bill')
        verbose_name_plural = _('Bills')
        indexes = [
            models.Index(fields=['bill_status']),
            models.Index(fields=['terms']),

            models.Index(fields=['cash_account']),
            models.Index(fields=['prepaid_account']),
            models.Index(fields=['unearned_account']),

            models.Index(fields=['date_due']),
            models.Index(fields=['date_draft']),
            models.Index(fields=['date_in_review']),
            models.Index(fields=['date_approved']),
            models.Index(fields=['date_paid']),
            models.Index(fields=['date_canceled']),
            models.Index(fields=['date_void']),

            models.Index(fields=['vendor']),
            models.Index(fields=['bill_number']),
        ]

    def __str__(self):
        return f'Bill: {self.bill_number} | {self.get_bill_status_display()}'

    def is_configured(self) -> bool:
        return all([
            super().is_configured(),
            self.bill_status
        ])

    # Configuration...
    def configure(self,
                  entity_slug: Union[str, EntityModel],
                  user_model: Optional[UserModel] = None,
                  date_draft: Optional[Union[date, datetime]] = None,
                  ledger_posted: bool = False,
                  ledger_name: str = None,
                  commit: bool = False,
                  commit_ledger: bool = False):
        """
        A configuration hook which executes all initial BillModel setup on to the LedgerModel and all initial
        values of the BillModel. Can only call this method once in the lifetime of a BillModel.

        Parameters
        ----------

        date_draft: date
            Optional date to use as Draft Date. Defaults to localdate() if None.
        entity_slug: str or EntityModel
            The entity slug or EntityModel to associate the Bill with.
        user_model: UserModel
            The UserModel making the request to check for QuerySet permissions.
        ledger_posted: bool
            An option to mark the BillModel Ledger as posted at the time of configuration. Defaults to False.
        ledger_name: str
            Optional additional InvoiceModel ledger name or description.
        commit: bool
            Saves the current BillModel after being configured.
        commit_ledger: bool
            Saves the BillModel's LedgerModel while being configured.

        Returns
        -------
        A tuple of LedgerModel, BillModel
        """

        # todo: add raise_exception flag, check if this is consistent...

        if not self.is_configured():
            if isinstance(entity_slug, str):
                if not user_model:
                    raise BillModelValidationError(_('Must pass user_model when using entity_slug.'))
                entity_qs = EntityModel.objects.for_user(user_model=user_model)
                entity_model: EntityModel = get_object_or_404(entity_qs, slug__exact=entity_slug)
            elif isinstance(entity_slug, EntityModel):
                entity_model = entity_slug
            else:
                raise BillModelValidationError('entity_slug must be an instance of str or EntityModel')

            if entity_model.is_accrual_method():
                self.accrue = True
                self.progress = Decimal.from_float(1.00)
            else:
                self.accrue = False

            self.bill_status = self.BILL_STATUS_DRAFT

            if date_draft and isinstance(date_draft, datetime):
                date_draft = date_draft.date()
            self.date_draft = get_localdate() if not date_draft else date_draft

            LedgerModel = lazy_loader.get_ledger_model()
            ledger_model: LedgerModel = LedgerModel(entity=entity_model, posted=ledger_posted)
            ledger_name = f'Bill {self.uuid}' if not ledger_name else ledger_name
            ledger_model.name = ledger_name
            ledger_model.configure_for_wrapper_model(model_instance=self)
            ledger_model.clean()
            ledger_model.clean_fields()
            self.ledger = ledger_model

            if commit_ledger or commit:
                self.ledger.save()

            if self.can_generate_bill_number():
                self.generate_bill_number(commit=commit)
                ledger_model.ledger_xid = f'bill-{self.bill_number.lower()}-{str(ledger_model.entity_id)[-5:]}'
                ledger_model.save(update_fields=['ledger_xid'])

            self.clean()

            if commit:
                self.save()

        return self.ledger, self

    # ### ItemizeMixIn implementation START...
    def can_migrate_itemtxs(self) -> bool:
        return self.is_draft()

    def migrate_itemtxs(self, itemtxs: Dict, operation: str, commit: bool = False):
        itemtxs_batch = super().migrate_itemtxs(itemtxs=itemtxs, commit=commit, operation=operation)
        self.update_amount_due(itemtxs_qs=itemtxs_batch)
        self.get_state(commit=True)

        if commit:
            self.save(update_fields=['amount_due',
                                     'amount_receivable',
                                     'amount_unearned',
                                     'amount_earned',
                                     'updated'])
        return itemtxs_batch

    def get_item_model_qs(self) -> ItemModelQuerySet:
        return ItemModel.objects.filter(
            entity_id__exact=self.ledger.entity_id
        ).bills()

    def validate_itemtxs_qs(self, queryset: Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]):
        """
        Validates that the entire ItemTransactionModelQuerySet is bound to the BillModel.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet or list of ItemTransactionModel.
            ItemTransactionModelQuerySet to validate.
        """
        valid = all([
            i.bill_model_id == self.uuid for i in queryset
        ])
        if not valid:
            raise BillModelValidationError(f'Invalid queryset. All items must be assigned to Bill {self.uuid}')

    def get_itemtxs_data(self,
                         queryset: Optional[ItemTransactionModelQuerySet] = None,
                         aggregate_on_db: bool = False,
                         lazy_agg: bool = False) -> Tuple[ItemTransactionModelQuerySet, Dict]:
        """
        Fetches the BillModel Items and aggregates the QuerySet.

        Parameters
        ----------
        queryset:
            Optional pre-fetched ItemModelQueryset to use. Avoids additional DB query if provided.
        aggregate_on_db: bool
            If True, performs aggregation of ItemsTransactions in the DB resulting in one additional DB query.
        Returns
        -------
        A tuple: ItemTransactionModelQuerySet, dict
        """
        if not queryset:
            queryset = self.itemtransactionmodel_set.all().select_related(
                'item_model',
                'entity_unit',
                'po_model',
                'bill_model')
        else:
            self.validate_itemtxs_qs(queryset)

        if aggregate_on_db and isinstance(queryset, ItemTransactionModelQuerySet):
            return queryset, queryset.aggregate(
                total_amount__sum=Sum('total_amount'),
                total_items=Count('uuid')
            )
        return queryset, {
            'total_amount__sum': sum(i.total_amount for i in queryset),
            'total_items': len(queryset)
        } if not lazy_agg else None

    # ### ItemizeMixIn implementation END...

    def get_transaction_queryset(self, annotated: bool = False):
        """
        Fetches the TransactionModelQuerySet associated with the BillModel instance.
        """
        TransactionModel = lazy_loader.get_txs_model()
        transaction_model_qs = TransactionModel.objects.all().for_ledger(ledger_model=self.ledger_id)
        if annotated:
            return transaction_model_qs.with_annotated_details()
        return transaction_model_qs

    # State..
    def get_migrate_state_desc(self) -> str:
        """
        Description used when migrating transactions into the LedgerModel.

        Returns
        _______
        str
            Description as a string.
        """
        return f'Bill {self.bill_number} account adjustment.'

    def get_migration_data(self,
                           queryset: Optional[ItemTransactionModelQuerySet] = None) -> ItemTransactionModelQuerySet:
        """
        Fetches necessary item transaction data to perform a migration into the LedgerModel.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet
            Optional pre-fetched ItemModelTransactionQueryset to use. Avoids additional DB query if provided.
        """

        if not queryset:
            queryset = self.itemtransactionmodel_set.all()
        else:
            self.validate_itemtxs_qs(queryset)

        return queryset.order_by('item_model__expense_account__uuid',
                                 'entity_unit__uuid',
                                 'item_model__expense_account__balance_type').values(
            'item_model__expense_account__uuid',
            'item_model__inventory_account__uuid',
            'item_model__expense_account__balance_type',
            'item_model__inventory_account__balance_type',
            'entity_unit__slug',
            'entity_unit__uuid',
            'total_amount').annotate(
            account_unit_total=Sum('total_amount')
        )

    def update_amount_due(self, itemtxs_qs: Optional[
        Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]] = None) -> ItemTransactionModelQuerySet:
        """
        Updates the BillModel amount due.

        Parameters
        ----------
        itemtxs_qs: ItemTransactionModelQuerySet or list of ItemTransactionModel
            Optional pre-fetched ItemTransactionModelQuerySet. Avoids additional DB if provided.
            Queryset is validated if provided.

        Returns
        -------
        ItemTransactionModelQuerySet
            Newly fetched of previously fetched ItemTransactionModelQuerySet if provided.
        """
        itemtxs_qs, itemtxs_agg = self.get_itemtxs_data(queryset=itemtxs_qs)
        self.amount_due = round(itemtxs_agg['total_amount__sum'], 2)
        return itemtxs_qs

    def is_draft(self) -> bool:
        """
        Checks if the BillModel is in Draft status.

        Returns
        _______
        bool
            True if BillModel is Draft, else False.
        """
        return self.bill_status == self.BILL_STATUS_DRAFT

    def is_review(self) -> bool:
        """
        Checks if the BillModel is In Review status.

        Returns
        _______
        bool
            True if BillModel is in Review, else False.
        """
        return self.bill_status == self.BILL_STATUS_REVIEW

    def is_approved(self) -> bool:
        """
        Checks if the BillModel is in Approved status.

        Returns
        _______
        bool
            True if BillModel is Approved, else False.
        """
        return self.bill_status == self.BILL_STATUS_APPROVED

    def is_paid(self) -> bool:
        """
        Checks if the BillModel is in Paid status.

        Returns
        _______
        bool
            True if BillModel is Paid, else False.
        """
        return self.bill_status == self.BILL_STATUS_PAID

    def is_canceled(self) -> bool:
        """
        Checks if the BillModel is in Canceled status.

        Returns
        _______
        bool
            True if BillModel is Canceled, else False.
        """
        return self.bill_status == self.BILL_STATUS_CANCELED

    def is_active(self):
        """
        Checks if the BillModel has the potential to impact the books and produce financial statements status.

        Returns
        _______
        bool
            True if BillModel is Active, else False.
        """
        return any([
            self.is_paid(),
            self.is_approved(),
            self.is_void()
        ])

    def is_void(self) -> bool:
        """
        Checks if the BillModel is in Void status.

        Returns
        _______
        bool
            True if BillModel is Void, else False.
        """
        return self.bill_status == self.BILL_STATUS_VOID

    def is_past_due(self) -> bool:
        """
        Checks if the BillModel is past due.

        Returns
        -------
        bool
            True if BillModel is past due, else False.
        """
        if self.date_due and self.is_approved():
            return self.date_due < get_localdate()
        return False

    # Permissions....
    def can_draft(self) -> bool:
        """
        Checks if the BillModel can be marked as Draft.

        Returns
        -------
        bool
            True if BillModel can be marked as draft, else False.
        """
        return self.is_review()

    def can_review(self) -> bool:
        """
        Checks if the BillModel can be marked as In Review.

        Returns
        -------
        bool
            True if BillModel can be marked as in review, else False.
        """
        return all([
            self.is_configured(),
            self.is_draft()
        ])

    def can_approve(self) -> bool:
        """
        Checks if the BillModel can be marked as Approved.

        Returns
        -------
        bool
            True if BillModel can be marked as approved, else False.
        """
        return self.is_review()

    def can_pay(self) -> bool:
        """
        Checks if the BillModel can be marked as Paid.

        Returns
        -------
        bool
            True if BillModel can be marked as paid, else False.
        """
        return self.is_approved()

    # todo: check on invoice
    # todo: check invoice canceled workflow...
    # todo: check invoice card for canceled
    def can_delete(self) -> bool:
        """
        Checks if the BillModel can be deleted.

        Returns
        -------
        bool
            True if BillModel can be deleted, else False.
        """
        return any([
            self.is_review(),
            self.is_draft(),
            not self.ledger.is_locked()
        ])

    def can_void(self) -> bool:
        """
        Checks if the BillModel can be marked as Void status.

        Returns
        -------
        bool
            True if BillModel can be marked as void, else False.
        """
        return all([
            self.is_approved(),
            float(self.amount_paid) == 0.00
        ])

    def can_cancel(self) -> bool:
        """
        Checks if the BillModel can be marked as Canceled status.

        Returns
        -------
        bool
            True if BillModel can be marked as canceled, else False.
        """
        return any([
            self.is_draft(),
            self.is_review()
        ])

    def can_edit_items(self) -> bool:
        """
        Checks if the BillModel item transactions can be edited.

        Returns
        -------
        bool
            True if BillModel items can be edited, else False.
        """
        return self.is_draft()

    def can_migrate(self) -> bool:
        """
        Checks if the BillModel can be migrated.

        Returns
        -------
        bool
            True if BillModel can be migrated, else False.
        """
        can_migrate = super().can_migrate()
        if not can_migrate:
            return False
        return self.is_approved()

    def can_bind_estimate(self, estimate_model, raise_exception: bool = False) -> bool:
        """
        Checks if the BillModel can be bound to a given EstimateModel.

        Parameters
        __________

        estimate_model: EstimateModel
            EstimateModel to check against.

        raise_exception: bool
            If True, raises BillModelValidationError if unable to bind. Else, returns False.

        Returns
        _______

        bool
            True if can bind provided EstimateModel, else False.
        """
        if self.ce_model_id:
            if raise_exception:
                raise BillModelValidationError(f'Bill {self.bill_number} already bound to '
                                               f'Estimate {self.ce_model.estimate_number}')
            return False

        is_approved = estimate_model.is_approved()
        if not is_approved and raise_exception:
            raise BillModelValidationError(f'Cannot bind estimate that is not approved.')
        return all([
            is_approved
        ])

    def can_bind_po(self, po_model, raise_exception: bool = False) -> bool:
        """
        Checks if the BillModel can be bound to a given PurchaseOrderModel.

        Parameters
        __________

        po_model: PurchaseOrderModel
            The PurchaseOrderModel to check against.

        raise_exception: bool
            If True, raises BillModelValidationError if unable to bind, else False.

        Returns
        _______

        bool
            True if can bind provided PurchaseOderModel, else False.
        """
        if not po_model.is_approved():
            if raise_exception:
                raise BillModelValidationError(f'Cannot bind an unapproved PO.')
            return False

        if po_model.date_approved > self.date_draft:
            if raise_exception:
                raise BillModelValidationError(f'Approved PO date cannot be greater than Bill draft date.')
            return False

        return True

    def can_generate_bill_number(self) -> bool:
        """
        Checks if BillModel can generate its Document Number.

        Returns
        _______

        bool
            True if BillModel can generate its bill_number, else False.
        """
        return all([
            not self.bill_number,
            self.is_draft(),
            self.is_configured()
        ])

    # ACTIONS ---

    def can_make_payment(self) -> bool:
        """
        Checks if the BillModel can accept a payment.


        Returns
        _______

        bool
            True if can bind provided PurchaseOderModel, else False.
        """
        return self.is_approved()

    def make_payment(self,
                     payment_amount: Union[Decimal, float, int],
                     payment_date: Optional[Union[datetime, date]] = None,
                     commit: bool = False,
                     raise_exception: bool = True):
        """
        Makes a payment to the BillModel.


        Parameters
        __________

        payment_amount: Decimal ot float
            The payment amount to process.

        payment_date: datetime or date.
            Date or timestamp of the payment being applied.

        commit: bool
            If True, commits the transaction into the DB. Defaults to False.

        raise_exception: bool
            If True, raises BillModelValidationError if payment exceeds amount due, else False.

        Returns
        _______

        bool
            True if can make payment, else False.
        """

        if isinstance(payment_amount, float):
            payment_amount = Decimal.from_float(payment_amount)
        elif isinstance(payment_amount, int):
            payment_amount = Decimal.from_float(float(payment_amount))
        self.amount_paid += payment_amount

        if self.amount_paid > self.amount_due:
            if raise_exception:
                raise BillModelValidationError(
                    f'Amount paid: {self.amount_paid} exceed amount due: {self.amount_due}.'
                )
            return

        self.get_state(commit=True)
        self.clean()

        if not payment_date:
            payment_date = get_localtime()

        if commit:
            self.migrate_state(
                user_model=None,
                entity_slug=self.ledger.entity.slug,
                je_timestamp=payment_date,
                raise_exception=True
            )
            self.save(
                update_fields=[
                    'amount_paid',
                    'amount_earned',
                    'amount_unearned',
                    'amount_receivable',
                    'updated'
                ])

    def bind_estimate(self, estimate_model, commit: bool = False, raise_exception: bool = True):
        """
        Binds BillModel to a given EstimateModel. Raises ValueError if EstimateModel cannot be bound.

        Parameters
        __________
        estimate_model: EstimateModel
            EstimateModel to bind.

        raise_exception: bool
            Raises BillModelValidationError if unable to bind EstimateModel.

        commit: bool
            Commits transaction into current BillModel.
        """
        try:
            self.can_bind_estimate(estimate_model, raise_exception=True)
        except ValueError as e:
            if raise_exception:
                raise e
            return
        self.ce_model = estimate_model
        self.clean()
        if commit:
            self.save(update_fields=[
                'ce_model',
                'updated'
            ])

    def mark_as_draft(self, date_draft: Optional[date] = None, commit: bool = False, **kwargs):
        """
        Marks BillModel as Draft.

        Parameters
        __________

        date_draft: date
            Draft date. If None, defaults to localdate().

        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_draft():
            raise BillModelValidationError(
                f'Bill {self.bill_number} cannot be marked as draft. Must be In Review.'
            )
        self.bill_status = self.BILL_STATUS_DRAFT

        if date_draft:
            if isinstance(date_draft, datetime):
                self.date_draft = date_draft.date()
            elif isinstance(date_draft, date):
                self.date_draft = date_draft
        else:
            self.date_draft = get_localdate()

        self.clean()
        if commit:
            self.save(
                update_fields=[
                    'bill_status',
                    'date_draft',
                    'updated'
                ]
            )
        bill_status_draft.send_robust(sender=self.__class__,
                                      instance=self,
                                      commited=commit, **kwargs)

    def get_mark_as_draft_html_id(self) -> str:
        """
        BillModel Mark as Draft HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-bill-model-{self.uuid}-mark-as-draft'

    def get_mark_as_draft_url(self, entity_slug: Optional[str] = None) -> str:
        """
        BillModel Mark-as-Draft action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            HTML ID as a String
        """
        if not entity_slug:
            entity_slug = self.ledger.entity.slug
        return reverse('django_ledger:bill-action-mark-as-draft',
                       kwargs={
                           'entity_slug': entity_slug,
                           'bill_pk': self.uuid
                       })

    def get_mark_as_draft_message(self) -> str:
        """
        Internationalized confirmation message with Bill Number.

        Returns
        _______
        str
            Mark-as-Draft BillModel confirmation message as a String.
        """
        return _('Do you want to mark Bill %s as Draft?') % self.bill_number

    # IN REVIEW ACTIONS....
    def mark_as_review(self,
                       commit: bool = False,
                       itemtxs_qs: ItemTransactionModelQuerySet = None,
                       date_in_review: Optional[date] = None,
                       raise_exception: bool = True,
                       **kwargs):
        """
        Marks BillModel as In Review.

        Parameters
        __________

        date_in_review: date
            BillModel in review date. Defaults to localdate() if None.
        itemtxs_qs: ItemTransactionModelQuerySet
            Pre fetched ItemTransactionModelQuerySet to use. Avoids additional DB Query if previously fetched.
        commit: bool
            Commits transaction into the Database. Defaults to False.
        raise_exception: bool
            Raises BillModelValidationError if BillModel cannot be marked as in review. Defaults to True.
        """
        if not self.can_review():
            if raise_exception:
                raise BillModelValidationError(
                    f'Bill {self.bill_number} cannot be marked as in review. Must be Draft and Configured.'
                )

        if not itemtxs_qs:
            itemtxs_qs = self.itemtransactionmodel_set.all()
        else:
            self.validate_itemtxs_qs(queryset=itemtxs_qs)

        if not itemtxs_qs.count():
            raise BillModelValidationError(message=f'Cannot review a {self.__class__.__name__} without items...')

        if not self.amount_due:
            raise BillModelValidationError(
                f'Bill {self.bill_number} cannot be marked as in review. Amount due must be greater than 0.'
            )

        self.bill_status = self.BILL_STATUS_REVIEW

        if date_in_review:
            if isinstance(date_in_review, datetime):
                self.date_in_review = date_in_review.date()
            elif isinstance(date_in_review, date):
                self.date_in_review = date_in_review
        else:
            self.date_in_review = get_localdate()

        self.clean()
        if commit:
            self.save(
                update_fields=[
                    'date_in_review',
                    'bill_status',
                    'updated'
                ]
            )
        bill_status_in_review.send_robust(sender=self.__class__,
                                          instance=self,
                                          commited=commit, **kwargs)

    def get_mark_as_review_html_id(self) -> str:
        """
        BillModel Mark as In Review HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}-mark-as-review'

    def get_mark_as_review_url(self, entity_slug: Optional[str] = None) -> str:
        """
        BillModel Mark-as-Review action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            BillModel mark-as-review action URL.
        """
        if not entity_slug:
            entity_slug = self.ledger.entity.slug
        return reverse('django_ledger:bill-action-mark-as-review',
                       kwargs={
                           'entity_slug': entity_slug,
                           'bill_pk': self.uuid
                       })

    def get_mark_as_review_message(self) -> str:
        """
        Internationalized confirmation message with Bill Number.

        Returns
        _______
        str
            Mark-as-Review BillModel confirmation message as a String.
        """
        return _('Do you want to mark Bill %s as In Review?') % self.bill_number

    # APPROVED ACTIONS....
    def mark_as_approved(self,
                         user_model,
                         entity_slug: Optional[str] = None,
                         date_approved: Optional[Union[date, datetime]] = None,
                         commit: bool = False,
                         force_migrate: bool = False,
                         raise_exception: bool = True,
                         **kwargs):
        """
        Marks BillModel as Approved.

        Parameters
        __________

        entity_slug
            Entity slug associated with the BillModel. Avoids additional DB query if passed.

        user_model
            UserModel associated with request.

        date_approved: date
            BillModel approved date. Defaults to localdate().

        commit: bool
            Commits transaction into the Database. Defaults to False.

        force_migrate: bool
            Forces migration. True if Accounting Method is Accrual.
        """
        if not self.can_approve():
            if raise_exception:
                raise BillModelValidationError(
                    f'Bill {self.bill_number} cannot be marked as in approved.'
                )
            return
        self.bill_status = self.BILL_STATUS_APPROVED

        if date_approved:
            if isinstance(date_approved, datetime):
                self.date_approved = date_approved.date()
            elif isinstance(date_approved, date):
                self.date_approved = date_approved
        else:
            self.date_approved = get_localdate()

        self.get_state(commit=True)
        self.clean()
        if commit:
            self.save()
            if force_migrate or self.accrue:
                if not entity_slug:
                    entity_slug = self.ledger.entity.slug
                # normally no transactions will be present when marked as approved...
                self.migrate_state(
                    entity_slug=entity_slug,
                    user_model=user_model,
                    je_timestamp=date_approved,
                    force_migrate=self.accrue
                )
            self.ledger.post(commit=commit, raise_exception=raise_exception)
        bill_status_approved.send_robust(sender=self.__class__,
                                         instance=self,
                                         commited=commit, **kwargs)

    def get_mark_as_approved_html_id(self) -> str:
        """
        BillModel Mark as Approved HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}-mark-as-approved'

    def get_mark_as_approved_url(self, entity_slug: Optional[str] = None) -> str:
        """
        BillModel Mark-as-Approved action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            BillModel mark-as-approved action URL.
        """
        if not entity_slug:
            entity_slug = self.ledger.entity.slug
        return reverse('django_ledger:bill-action-mark-as-approved',
                       kwargs={
                           'entity_slug': entity_slug,
                           'bill_pk': self.uuid
                       })

    def get_mark_as_approved_message(self) -> str:
        """
        Internationalized confirmation message with Bill Number.

        Returns
        _______
        str
            Mark-as-Approved BillModel confirmation message as a String.
        """
        return _('Do you want to mark Bill %s as Approved?') % self.bill_number

    # PAY ACTIONS....
    def mark_as_paid(self,
                     user_model,
                     entity_slug: Optional[str] = None,
                     date_paid: Optional[Union[date, datetime]] = None,
                     itemtxs_qs: Optional[ItemTransactionModelQuerySet] = None,
                     commit: bool = False,
                     **kwargs):

        """
        Marks BillModel as Paid.

        Parameters
        __________

        entity_slug: str
            Entity slug associated with the BillModel. Avoids additional DB query if passed.

        user_model:
            UserModel associated with request.

        date_paid: date
            BillModel paid date. Defaults to localdate() if None.

        itemtxs_qs: ItemTransactionModelQuerySet
            Pre-fetched ItemTransactionModelQuerySet. Avoids additional DB query. Validated if passed.

        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_pay():
            raise BillModelValidationError(f'Cannot mark Bill {self.bill_number} as paid...')

        if date_paid:
            if isinstance(date_paid, datetime):
                self.date_paid = date_paid.date()
            elif isinstance(date_paid, date):
                self.date_paid = date_paid
        else:
            self.date_paid = get_localdate()

        self.progress = Decimal.from_float(1.0)
        self.amount_paid = self.amount_due

        if self.date_paid > get_localdate():
            raise BillModelValidationError(f'Cannot pay {self.__class__.__name__} in the future.')
        if self.date_paid < self.date_approved:
            raise BillModelValidationError(
                f'Cannot pay {self.__class__.__name__} before approved date {self.date_approved}.')

        self.bill_status = self.BILL_STATUS_PAID
        self.get_state(commit=True)
        self.clean()

        if not itemtxs_qs:
            itemtxs_qs = self.itemtransactionmodel_set.all()
        else:
            self.validate_itemtxs_qs(queryset=itemtxs_qs)

        if commit:
            self.save()
            ItemTransactionModel = lazy_loader.get_item_transaction_model()
            itemtxs_qs.filter(
                po_model_id__isnull=False
            ).update(po_item_status=ItemTransactionModel.STATUS_ORDERED)

            if not entity_slug:
                entity_slug = self.ledger.entity.slug

            self.migrate_state(
                user_model=user_model,
                entity_slug=entity_slug,
                itemtxs_qs=itemtxs_qs,
                je_timestamp=date_paid,
                force_migrate=True
            )
            self.lock_ledger(commit=True)
        bill_status_paid.send_robust(sender=self.__class__,
                                     instance=self,
                                     commited=commit, **kwargs)

    def get_mark_as_paid_html_id(self) -> str:
        """
        BillModel Mark as Paid HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-bill-model-{self.uuid}-mark-as-paid'

    def get_mark_as_paid_url(self, entity_slug: Optional[str] = None) -> str:
        """
        BillModel Mark-as-Paid action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            BillModel mark-as-paid action URL.
        """
        if not entity_slug:
            entity_slug = self.ledger.entity.slug
        return reverse('django_ledger:bill-action-mark-as-paid',
                       kwargs={
                           'entity_slug': entity_slug,
                           'bill_pk': self.uuid
                       })

    def get_mark_as_paid_message(self) -> str:
        """
        Internationalized confirmation message with Bill Number.

        Returns
        _______
        str
            Mark-as-Paid BillModel confirmation message as a String.
        """
        return _('Do you want to mark Bill %s as Paid?') % self.bill_number

    # VOID Actions...
    def mark_as_void(self,
                     user_model,
                     entity_slug: Optional[str] = None,
                     date_void: Optional[date] = None,
                     commit: bool = False,
                     **kwargs):
        """
        Marks BillModel as Void.
        When mark as void, all transactions associated with BillModel are reversed as of the void date.

        Parameters
        __________

        entity_slug: str
            Entity slug associated with the BillModel. Avoids additional DB query if passed.

        user_model
            UserModel associated with request.

        date_void: date
            BillModel void date. Defaults to localdate() if None.

        commit: bool
            Commits transaction into DB. Defaults to False.
        """
        if not self.can_void():
            raise BillModelValidationError(f'Bill {self.bill_number} cannot be voided. Must be approved.')

        if date_void:
            if isinstance(date_void, datetime):
                self.date_void = date_void.date()
            elif isinstance(date_void, date):
                self.date_void = date_void
        else:
            self.date_void = get_localdate()

        self.bill_status = self.BILL_STATUS_VOID
        self.void_state(commit=True)
        self.clean()

        if commit:
            if not entity_slug:
                entity_slug = self.ledger.entity.slug

            self.unlock_ledger(commit=False, raise_exception=False)
            self.migrate_state(
                entity_slug=entity_slug,
                user_model=user_model,
                void=True,
                void_date=self.date_void,
                raise_exception=False,
                force_migrate=True)
            self.save()
            self.lock_ledger(commit=False, raise_exception=False)
        bill_status_void.send_robust(sender=self.__class__,
                                     instance=self,
                                     commited=commit, **kwargs)

    def get_mark_as_void_html_id(self) -> str:
        """
        BillModel Mark as Void HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}-mark-as-void'

    def get_mark_as_void_url(self, entity_slug: Optional[str] = None) -> str:
        """
        BillModel Mark-as-Void action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
            BillModel mark-as-void action URL.
        """
        if not entity_slug:
            entity_slug = self.ledger.entity.slug
        return reverse('django_ledger:bill-action-mark-as-void',
                       kwargs={
                           'entity_slug': entity_slug,
                           'bill_pk': self.uuid
                       })

    def get_mark_as_void_message(self) -> str:
        """
        Internationalized confirmation message with Bill Number.

        Returns
        _______
        str
            Mark-as-Void BillModel confirmation message as a String.
        """
        return _('Do you want to void Bill %s?') % self.bill_number

    # Cancel Actions...
    def mark_as_canceled(self, date_canceled: Optional[date] = None, commit: bool = False, **kwargs):
        """
        Mark BillModel as Canceled.

        Parameters
        __________

        date_canceled: date
            BillModel canceled date. Defaults to localdate() if None.

        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_cancel():
            raise BillModelValidationError(f'Bill {self.bill_number} cannot be canceled. Must be draft or in review.')

        self.date_canceled = get_localdate() if not date_canceled else date_canceled
        self.bill_status = self.BILL_STATUS_CANCELED
        self.clean()
        if commit:
            self.save()
        bill_status_canceled.send_robust(sender=self.__class__,
                                         instance=self,
                                         commited=commit, **kwargs)

    def get_mark_as_canceled_html_id(self) -> str:
        """
        BillModel Mark as Canceled HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}-mark-as-canceled'

    def get_mark_as_canceled_url(self, entity_slug: Optional[str] = None) -> str:
        """
        BillModel Mark-as-Canceled action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            BillModel mark-as-canceled action URL.
        """

        if not entity_slug:
            entity_slug = self.ledger.entity.slug

        return reverse('django_ledger:bill-action-mark-as-canceled',
                       kwargs={
                           'entity_slug': entity_slug,
                           'bill_pk': self.uuid
                       })

    def get_mark_as_canceled_message(self) -> str:
        """
        Internationalized confirmation message with Bill Number.

        Returns
        _______
        str
            Mark-as-Canceled BillModel confirmation message as a String.
        """
        return _('Do you want to mark Bill %s as Canceled?') % self.bill_number

    # DELETE ACTIONS...
    def delete(self, force_db_delete: bool = False, using=None, keep_parents=False):
        if not force_db_delete:
            self.mark_as_canceled(commit=True)
            return
        if not self.can_delete():
            raise BillModelValidationError(
                message=_(f'Bill {self.bill_number} cannot be deleted...')
            )
        return super().delete(using=using, keep_parents=keep_parents)

    # def mark_as_delete(self, **kwargs):
    #     """
    #     Deletes BillModel from DB if possible. Raises exception if can_delete() is False.
    #     """
    #     return self.delete()
    #
    # def get_mark_as_delete_html_id(self) -> str:
    #     """
    #     BillModel Mark as Delete HTML ID Tag.
    #
    #     Returns
    #     _______
    #
    #     str
    #         HTML ID as a String.
    #     """
    #     return f'djl-bill-model-{self.uuid}-mark-as-delete'
    #
    # def get_mark_as_delete_url(self, entity_slug: Optional[str] = None) -> str:
    #     """
    #     BillModel Mark-as-Delete action URL.
    #
    #     Parameters
    #     __________
    #     entity_slug: str
    #         Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
    #         is not cached on QuerySet.
    #
    #     Returns
    #     _______
    #     str
    #         BillModel mark-as-delete action URL.
    #     """
    #     if not entity_slug:
    #         entity_slug = self.ledger.entity.slug
    #     return reverse('django_ledger:bill-action-mark-as-delete',
    #                    kwargs={
    #                        'entity_slug': entity_slug,
    #                        'bill_pk': self.uuid
    #                    })
    #
    # def get_mark_as_delete_message(self) -> str:
    #     """
    #     Internationalized confirmation message with Bill Number.
    #
    #     Returns
    #     _______
    #     str
    #         Mark-as-Delete BillModel confirmation message as a String.
    #     """
    #     return _('Do you want to delete Bill %s?') % self.bill_number

    def get_status_action_date(self) -> date:
        """
        Current status action date.

        Returns
        _______
        date
            A date. i.e. If status is Approved, return date_approved. If Paid, return date_paid.
        """
        return getattr(self, f'date_{self.bill_status}')

    # HTML Tags...
    def get_document_id(self) -> Optional[str]:
        """
        Human-readable document number. Defaults to bill_number.

        Returns
        _______
        str
            Document Number as a String.
        """
        return self.bill_number

    def get_html_id(self) -> str:
        """
        Unique BillNumber HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}'

    def get_html_amount_due_id(self) -> str:
        """
        Unique amount due HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}-amount-due'

    def get_html_amount_paid_id(self) -> str:
        """
        Unique amount paid HTML ID

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}-amount-paid'

    def get_html_form_id(self) -> str:
        """
        Unique BillModel Form HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-bill-model-{self.uuid}-form'

    def get_terms_start_date(self) -> Optional[date]:
        """
        Date where BillModel term start to apply.

        Returns
        _______
        date
            A date which represents the start of BillModel terms.
        """
        return self.date_approved

    def _get_next_state_model(self, raise_exception: bool = True):
        """
        Fetches the next sequenced state model associated with the BillModel number.

        Parameters
        ----------
        raise_exception: bool
            Raises IntegrityError if unable to secure transaction from DB.

        Returns
        -------
        EntityStateModel
            An instance of EntityStateModel
        """
        EntityStateModel = lazy_loader.get_entity_state_model()
        EntityModel = lazy_loader.get_entity_model()
        entity_model = EntityModel.objects.get(uuid__exact=self.ledger.entity_id)
        fy_key = entity_model.get_fy_for_date(dt=self.date_draft)

        try:
            LOOKUP = {
                'entity_model_id__exact': self.ledger.entity_id,
                'entity_unit_id__exact': None,
                'fiscal_year': fy_key,
                'key__exact': EntityStateModel.KEY_BILL
            }

            state_model_qs = EntityStateModel.objects.filter(**LOOKUP).select_related(
                'entity_model').select_for_update()
            state_model = state_model_qs.get()
            state_model.sequence = F('sequence') + 1
            state_model.save(update_fields=['sequence'])
            state_model.refresh_from_db()
            return state_model
        except ObjectDoesNotExist:
            EntityModel = lazy_loader.get_entity_model()
            entity_model = EntityModel.objects.get(uuid__exact=self.ledger.entity_id)
            fy_key = entity_model.get_fy_for_date(dt=self.date_draft)

            LOOKUP = {
                'entity_model_id': entity_model.uuid,
                'entity_unit_id': None,
                'fiscal_year': fy_key,
                'key': EntityStateModel.KEY_BILL,
                'sequence': 1
            }
            state_model = EntityStateModel.objects.create(**LOOKUP)
            return state_model
        except IntegrityError as e:
            if raise_exception:
                raise e

    def generate_bill_number(self, commit: bool = False) -> str:
        """
        Atomic Transaction. Generates the next BillModel document number available. The operation
        will result in two additional queries if the BillModel & LedgerModel is not cached in
        QuerySet via select_related('ledger').

        Parameters
        __________
        commit: bool
            Commits transaction into BillModel.

        Returns
        _______
        str
            A String, representing the generated BillModel instance Document Number.
        """
        if self.can_generate_bill_number():
            with transaction.atomic(durable=True):

                state_model = None
                while not state_model:
                    state_model = self._get_next_state_model(raise_exception=False)

                seq = str(state_model.sequence).zfill(DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING)
                self.bill_number = f'{DJANGO_LEDGER_BILL_NUMBER_PREFIX}-{state_model.fiscal_year}-{seq}'

                if commit:
                    self.save(update_fields=['bill_number', 'updated'])

        return self.bill_number

    def generate_descriptive_title(self) -> str:
        return f'Bill {self.bill_number} | {self.get_bill_status_display()} {self.get_status_action_date()} | {self.vendor.vendor_name}'

    # --> URLs <---
    def get_absolute_url(self):
        return reverse('django_ledger:bill-detail',
                       kwargs={
                           'entity_slug': self.ledger.entity.slug,
                           'bill_pk': self.uuid
                       })

    def clean(self, commit: bool = True):
        """
        Clean method for BillModel. Results in a DB query if bill number has not been generated and the BillModel is
        eligible to generate a bill_number.

        Parameters
        __________

        commit: bool
            If True, commits into DB the generated BillModel number if generated.
        """

        super().clean()
        if self.cash_account.role != ASSET_CA_CASH:
            raise ValidationError(f'Cash account must be of role {ASSET_CA_CASH}.')
        if self.prepaid_account.role != ASSET_CA_PREPAID:
            raise ValidationError(f'Prepaid account must be of role {ASSET_CA_PREPAID}.')
        if self.unearned_account.role != LIABILITY_CL_ACC_PAYABLE:
            raise ValidationError(f'Unearned account must be of role {LIABILITY_CL_ACC_PAYABLE}.')


class BillModel(BillModelAbstract):
    """
    Base BillModel from Abstract.
    """

    class Meta(BillModelAbstract.Meta):
        abstract = False


def billmodel_presave(instance: BillModel, **kwargs):
    if instance.can_generate_bill_number():
        instance.generate_bill_number(commit=False)

    if not instance.entity_model_id:
        instance.entity_model = instance.ledger.entity


pre_save.connect(receiver=billmodel_presave, sender=BillModel)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.http import HttpResponseRedirect, HttpResponseNotFound, HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import (UpdateView, CreateView, DeleteView, MonthArchiveView,
                                  ArchiveIndexView, YearArchiveView, DetailView, RedirectView)
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.invoice import (BaseInvoiceModelUpdateForm, InvoiceModelCreateForEstimateForm,
                                         get_invoice_itemtxs_formset_class,
                                         DraftInvoiceModelUpdateForm, InReviewInvoiceModelUpdateForm,
                                         ApprovedInvoiceModelUpdateForm, PaidInvoiceModelUpdateForm,
                                         AccruedAndApprovedInvoiceModelUpdateForm, InvoiceModelCreateForm)
from django_ledger.io.io_core import get_localdate
from django_ledger.models import EntityModel, LedgerModel, EstimateModel
from django_ledger.models.invoice import InvoiceModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class InvoiceModelModelViewQuerySetMixIn:
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = InvoiceModel.objects.for_entity(
                entity_model=self.kwargs['entity_slug']
            ).select_related('customer', 'ledger').order_by('-created')
        return super().get_queryset()


class InvoiceModelListView(DjangoLedgerSecurityMixIn, InvoiceModelModelViewQuerySetMixIn, ArchiveIndexView):
    template_name = 'django_ledger/invoice/invoice_list.html'
    context_object_name = 'invoice_list'
    PAGE_TITLE = _('Invoice List')
    date_field = 'created'
    paginate_by = 20
    paginate_orphans = 2
    allow_empty = True
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }


class InvoiceModelYearlyListView(YearArchiveView, InvoiceModelListView):
    paginate_by = 10
    make_object_list = True


class InvoiceModelMonthlyListView(MonthArchiveView, InvoiceModelListView):
    paginate_by = 10
    month_format = '%m'


class InvoiceModelCreateView(DjangoLedgerSecurityMixIn, InvoiceModelModelViewQuerySetMixIn, CreateView):
    template_name = 'django_ledger/invoice/invoice_create.html'
    PAGE_TITLE = _('Create Invoice')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }
    for_estimate = False

    def get(self, request, entity_slug, **kwargs):

        if not request.user.is_authenticated:
            return HttpResponseForbidden()

        if self.for_estimate and 'ce_pk' in self.kwargs:
            estimate_qs = EstimateModel.objects.for_entity(
                entity_slug=entity_slug,
                user_model=self.request.user
            )
            estimate_model: EstimateModel = get_object_or_404(estimate_qs, uuid__exact=self.kwargs['ce_pk'])
            if not estimate_model.can_bind():
                return HttpResponseNotFound('404 Not Found')
        return super(InvoiceModelCreateView, self).get(request, entity_slug, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(InvoiceModelCreateView, self).get_context_data(**kwargs)

        if self.for_estimate:
            context['form_action_url'] = reverse('django_ledger:invoice-create-estimate',
                                                 kwargs={
                                                     'entity_slug': self.kwargs['entity_slug'],
                                                     'ce_pk': self.kwargs['ce_pk']
                                                 })
            estimate_qs = EstimateModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL,
            ).select_related('customer')
            estimate_model = get_object_or_404(estimate_qs, uuid__exact=self.kwargs['ce_pk'])
            context['estimate_model'] = estimate_model
        else:
            context['form_action_url'] = reverse('django_ledger:invoice-create',
                                                 kwargs={
                                                     'entity_slug': self.kwargs['entity_slug']
                                                 })
        return context

    def get_initial(self):
        return {
            'date_draft': get_localdate()
        }

    def get_form(self, form_class=None):
        entity_slug = self.kwargs['entity_slug']
        if self.for_estimate:
            InvoiceModelCreateForEstimateForm(
                entity_slug=entity_slug,
                user_model=self.request.user,
                **self.get_form_kwargs()
            )
        return InvoiceModelCreateForm(
            entity_slug=entity_slug,
            user_model=self.request.user,
            **self.get_form_kwargs()
        )

    def form_valid(self, form):
        invoice_model: InvoiceModel = form.save(commit=False)
        ledger_model, invoice_model = invoice_model.configure(
            entity_slug=self.AUTHORIZED_ENTITY_MODEL,
            commit_ledger=True
        )

        if self.for_estimate:
            ce_pk = self.kwargs['ce_pk']
            estimate_model_qs = EstimateModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL,
            )

            estimate_model = get_object_or_404(estimate_model_qs, uuid__exact=ce_pk)
            invoice_model.bind_estimate(estimate_model=estimate_model, commit=False)

        return super().form_valid(form=form)

    def get_success_url(self):
        entity_slug = self.kwargs['entity_slug']
        invoice_model: InvoiceModel = self.object
        if self.for_estimate:
            return reverse('django_ledger:customer-estimate-detail',
                           kwargs={
                               'entity_slug': entity_slug,
                               'ce_pk': self.kwargs['ce_pk']
                           })
        return reverse('django_ledger:invoice-detail',
                       kwargs={
                           'entity_slug': entity_slug,
                           'invoice_pk': invoice_model.uuid
                       })


class InvoiceModelUpdateView(DjangoLedgerSecurityMixIn, InvoiceModelModelViewQuerySetMixIn, UpdateView):
    slug_url_kwarg = 'invoice_pk'
    slug_field = 'uuid'
    context_object_name = 'invoice'
    template_name = 'django_ledger/invoice/invoice_update.html'
    form_class = BaseInvoiceModelUpdateForm
    http_method_names = ['get', 'post']

    action_update_items = False

    def get_form_class(self):
        invoice_model: InvoiceModel = self.object

        if invoice_model.is_draft():
            return DraftInvoiceModelUpdateForm
        elif invoice_model.is_review():
            return InReviewInvoiceModelUpdateForm
        elif invoice_model.is_approved():
            if invoice_model.accrue:
                return AccruedAndApprovedInvoiceModelUpdateForm
            return ApprovedInvoiceModelUpdateForm
        elif invoice_model.is_paid():
            return PaidInvoiceModelUpdateForm
        return BaseInvoiceModelUpdateForm

    def get_form(self, form_class=None):
        form_class = self.get_form_class()
        if self.request.method == 'POST' and self.action_update_items:
            return form_class(
                entity_slug=self.kwargs['entity_slug'],
                user_model=self.request.user,
                instance=self.object
            )
        return form_class(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            **self.get_form_kwargs()
        )

    def get_context_data(self, itemtxs_formset=None, **kwargs):
        context = super().get_context_data(**kwargs)
        invoice_model: InvoiceModel = self.object
        title = f'Invoice {invoice_model.invoice_number}'
        context['page_title'] = title
        context['header_title'] = title

        ledger_model: LedgerModel = self.object.ledger

        if not invoice_model.is_configured():
            messages.add_message(
                request=self.request,
                message=f'Invoice {invoice_model.invoice_number} must have all accounts configured.',
                level=messages.ERROR,
                extra_tags='is-danger'
            )

        if not invoice_model.is_paid():
            if ledger_model.locked:
                messages.add_message(self.request,
                                     messages.ERROR,
                                     f'Warning! This invoice is locked. Must unlock before making any changes.',
                                     extra_tags='is-danger')

        if ledger_model.locked:
            messages.add_message(self.request,
                                 messages.ERROR,
                                 f'Warning! This Invoice is Locked. Must unlock before making any changes.',
                                 extra_tags='is-danger')

        if not ledger_model.is_posted():
            messages.add_message(self.request,
                                 messages.INFO,
                                 f'This Invoice has not been posted. Must post to see ledger changes.',
                                 extra_tags='is-info')

        if not itemtxs_formset:
            itemtxs_qs = invoice_model.itemtransactionmodel_set.all().select_related('item_model')
            itemtxs_qs, itemtxs_agg = invoice_model.get_itemtxs_data(queryset=itemtxs_qs)
            invoice_itemtxs_formset_class = get_invoice_itemtxs_formset_class(invoice_model)
            itemtxs_formset = invoice_itemtxs_formset_class(
                entity_slug=self.kwargs['entity_slug'],
                user_model=self.request.user,
                invoice_model=invoice_model,
                queryset=itemtxs_qs
            )
        else:
            itemtxs_qs, itemtxs_agg = invoice_model.get_itemtxs_data(queryset=itemtxs_formset.queryset)

        context['itemtxs_formset'] = itemtxs_formset
        context['total_amount__sum'] = itemtxs_agg['total_amount__sum']
        return context

    def get_success_url(self):
        entity_slug = self.kwargs['entity_slug']
        invoice_pk = self.kwargs['invoice_pk']
        return reverse('django_ledger:invoice-update',
                       kwargs={
                           'entity_slug': entity_slug,
                           'invoice_pk': invoice_pk
                       })

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.prefetch_related('itemtransactionmodel_set')

    def form_valid(self, form):
        invoice_model: InvoiceModel = form.save(commit=False)
        if invoice_model.can_migrate():
            invoice_model.migrate_state(
                user_model=self.request.user,
                entity_slug=self.kwargs['entity_slug']
            )
        messages.add_message(self.request,
                             messages.SUCCESS,
                             f'Invoice {self.object.invoice_number} successfully updated.',
                             extra_tags='is-success')
        return super().form_valid(form)

    def get(self, request, entity_slug, invoice_pk, *args, **kwargs):
        if self.action_update_items:
            return HttpResponseRedirect(
                redirect_to=reverse('django_ledger:invoice-update',
                                    kwargs={
                                        'entity_slug': entity_slug,
                                        'invoice_pk': invoice_pk
                                    })
            )
        return super(InvoiceModelUpdateView, self).get(request, *args, **kwargs)

    def post(self, request, entity_slug, invoice_pk, *args, **kwargs):
        if self.action_update_items:
            if not request.user.is_authenticated:
                return HttpResponseForbidden()

            queryset = self.get_queryset()
            invoice_model = self.get_object(queryset=queryset)
            self.object = invoice_model
            invoice_itemtxs_formset_class = get_invoice_itemtxs_formset_class(invoice_model)
            itemtxs_formset = invoice_itemtxs_formset_class(request.POST,
                                                            user_model=self.request.user,
                                                            invoice_model=invoice_model,
                                                            entity_slug=entity_slug)

            if not invoice_model.can_edit_items():
                messages.add_message(
                    request,
                    message=f'Cannot update items once Invoice is {invoice_model.get_invoice_status_display()}',
                    level=messages.ERROR,
                    extra_tags='is-danger'
                )
                context = self.get_context_data(itemtxs_formset=itemtxs_formset)
                return self.render_to_response(context=context)

            if itemtxs_formset.has_changed():
                if itemtxs_formset.is_valid():
                    itemtxs_list = itemtxs_formset.save(commit=False)
                    entity_qs = EntityModel.objects.for_user(user_model=self.request.user)
                    entity_model: EntityModel = get_object_or_404(entity_qs, slug__exact=entity_slug)

                    for itemtxs in itemtxs_list:
                        itemtxs.invoice_model_id = invoice_model.uuid
                        itemtxs.clean()

                    itemtxs_list = itemtxs_formset.save()
                    itemtxs_qs = invoice_model.update_amount_due()
                    invoice_model.get_state(commit=True)
                    invoice_model.clean()
                    invoice_model.save(
                        update_fields=['amount_due',
                                       'amount_receivable',
                                       'amount_unearned',
                                       'amount_earned',
                                       'updated']
                    )

                    invoice_model.migrate_state(
                        entity_slug=entity_slug,
                        user_model=self.request.user,
                        raise_exception=False,
                        itemtxs_qs=itemtxs_qs
                    )

                    messages.add_message(request,
                                         message=f'Items for Invoice {invoice_model.invoice_number} saved.',
                                         level=messages.SUCCESS,
                                         extra_tags='is-success')
                    return HttpResponseRedirect(
                        redirect_to=reverse('django_ledger:invoice-update',
                                            kwargs={
                                                'entity_slug': entity_slug,
                                                'invoice_pk': invoice_pk
                                            })
                    )

                # if not valid, return formset with errors...
                return self.render_to_response(context=self.get_context_data(itemtxs_formset=itemtxs_formset))
        return super(InvoiceModelUpdateView, self).post(request, **kwargs)


class InvoiceModelDetailView(DjangoLedgerSecurityMixIn, InvoiceModelModelViewQuerySetMixIn, DetailView):
    slug_url_kwarg = 'invoice_pk'
    slug_field = 'uuid'
    context_object_name = 'invoice'
    template_name = 'django_ledger/invoice/invoice_detail.html'
    extra_context = {
        'hide_menu': True
    }

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        invoice = self.object.invoice_number
        title = f'Invoice {invoice}'
        context['page_title'] = title
        context['header_title'] = title

        invoice_model: InvoiceModel = self.object
        itemtxs_qs, itemtxs_agg = invoice_model.get_itemtxs_data()
        context['itemtxs_qs'] = itemtxs_qs
        context['total_amount_due'] = itemtxs_agg['total_amount__sum']
        return context

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related(
            'ledger',
            'ledger__entity',
            'customer',
            'cash_account',
            'prepaid_account',
            'unearned_account',
            'cash_account__coa_model',
            'prepaid_account__coa_model',
            'unearned_account__coa_model'
        )


class InvoiceModelDeleteView(DjangoLedgerSecurityMixIn, InvoiceModelModelViewQuerySetMixIn, DeleteView):
    slug_url_kwarg = 'invoice_pk'
    slug_field = 'uuid'
    template_name = 'django_ledger/invoice/invoice_delete.html'
    context_object_name = 'invoice'
    extra_context = {
        'hide_menu': True
    }

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context['page_title'] = _('Delete Invoice ') + self.object.invoice_number
        context['header_title'] = context['page_title']
        return context

    def get_success_url(self):
        return reverse('django_ledger:entity-dashboard',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })


# ACTION VIEWS...
class BaseInvoiceActionView(DjangoLedgerSecurityMixIn,
                            RedirectView,
                            InvoiceModelModelViewQuerySetMixIn,
                            SingleObjectMixin):
    http_method_names = ['get']
    pk_url_kwarg = 'invoice_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, *args, **kwargs):
        return reverse('django_ledger:invoice-update',
                       kwargs={
                           'entity_slug': kwargs['entity_slug'],
                           'invoice_pk': kwargs['invoice_pk']
                       })

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(BaseInvoiceActionView, self).get(request, *args, **kwargs)
        invoice_model: InvoiceModel = self.get_object()

        try:
            getattr(invoice_model, self.action_name)(commit=self.commit, **kwargs)
        except ValidationError as e:
            messages.add_message(request,
                                 message=e.message,
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
        return response


class InvoiceModelActionMarkAsDraftView(BaseInvoiceActionView):
    action_name = 'mark_as_draft'


class InvoiceModelActionMarkAsReviewView(BaseInvoiceActionView):
    action_name = 'mark_as_review'


class InvoiceModelActionMarkAsApprovedView(BaseInvoiceActionView):
    action_name = 'mark_as_approved'


class InvoiceModelActionMarkAsPaidView(BaseInvoiceActionView):
    action_name = 'mark_as_paid'


class InvoiceModelActionMarkAsCanceledView(BaseInvoiceActionView):
    action_name = 'mark_as_canceled'


class InvoiceModelActionMarkAsVoidView(BaseInvoiceActionView):
    action_name = 'mark_as_void'


class InvoiceModelActionLockLedgerView(BaseInvoiceActionView):
    action_name = 'lock_ledger'


class InvoiceModelActionUnlockLedgerView(BaseInvoiceActionView):
    action_name = 'unlock_ledger'


class InvoiceModelActionForceMigrateView(BaseInvoiceActionView):
    action_name = 'migrate_state'

    def get_redirect_url(self, entity_slug, invoice_pk, *args, **kwargs):
        return reverse('django_ledger:invoice-update',
                       kwargs={
                           'entity_slug': entity_slug,
                           'invoice_pk': invoice_pk
                       })



from datetime import date, datetime
from random import choice, randint
from typing import Union, Optional, List
from urllib.parse import urlparse
from uuid import uuid4

from django.contrib.auth import get_user_model
from django.urls import reverse

from django_ledger.models import EntityModel, PurchaseOrderModel
from django_ledger.tests.base import DjangoLedgerBaseTest
from django_ledger.urls.purchase_order import urlpatterns as po_urls

UserModel = get_user_model()


class PurchaseOrderModelTests(DjangoLedgerBaseTest):

    def setUp(self) -> None:
        self.URL_PATTERNS = {
            p.name: set(p.pattern.converters.keys()) for p in po_urls
        }

    def create_purchase_order(self,
                              entity_model: EntityModel,
                              draft_date: Optional[Union[date, datetime]] = None) -> PurchaseOrderModel:
        po_model = PurchaseOrderModel()
        po_model = po_model.configure(entity_slug=entity_model,
                                      draft_date=draft_date,
                                      user_model=self.user_model,
                                      commit=True)
        return po_model

    def get_purchase_orders(self, entity_model: EntityModel) -> List[PurchaseOrderModel]:
        return PurchaseOrderModel.objects.for_entity(entity_model=entity_model)

    def test_protected_views(self):
        """
        All Purchase Model Views must have user authenticated.
        If not, user must be redirected to login page.
        """
        self.logout_client()

        entity_model: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)
        po_model = self.create_purchase_order(
            entity_model=entity_model,
            draft_date=self.get_random_draft_date()
        )

        self.assertEqual(len(self.URL_PATTERNS), 15)
        for path, kwargs in self.URL_PATTERNS.items():
            url_kwargs = dict()
            url_kwargs['entity_slug'] = entity_model.slug
            if 'po_pk' in kwargs:
                url_kwargs['po_pk'] = po_model.uuid
            if 'year' in kwargs:
                url_kwargs['year'] = po_model.date_draft.year
            if 'month' in kwargs:
                url_kwargs['month'] = po_model.date_draft.month
            if 'ce_pk' in kwargs:
                url_kwargs['ce_pk'] = uuid4()
            if 'po_pk' in kwargs:
                url_kwargs['po_pk'] = uuid4()

            url = reverse(f'django_ledger:{path}', kwargs=url_kwargs)
            redirect_response = self.CLIENT.get(url, follow=False)
            self.assertEqual(redirect_response.status_code, 302, msg=f'{path} view is not protected.')

    def test_purchase_order_list(self):
        """
        Check links and data shown on Purchase Order List page are as supposed to be.
        """
        self.login_client()
        entity_model: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)
        po_list_url = reverse('django_ledger:po-list', kwargs={'entity_slug': entity_model.slug})

        with self.assertNumQueries(6):  # previously 5
            response = self.CLIENT.get(po_list_url)
        self.assertEqual(response.status_code, 200, msg="Fail to GET Purchase Order list page")

        po_model_qs = response.context['po_list']

        for po_model in po_model_qs:
            po_detail_url = reverse('django_ledger:po-detail',
                                    kwargs={
                                        'entity_slug': entity_model.slug,
                                        'po_pk': po_model.uuid
                                    })

            po_delete_url = reverse('django_ledger:po-delete',
                                    kwargs={
                                        'entity_slug': entity_model.slug,
                                        'po_pk': po_model.uuid
                                    })
            # contains po-detail and po-delete urls
            self.assertContains(response, po_detail_url)
            self.assertContains(response, po_delete_url)

    def test_purchase_order_create(self):
        """
        Check Purchase Order create-page.
        """
        self.login_client()
        entity_model: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)

        po_create_url = reverse('django_ledger:po-create', kwargs={'entity_slug': entity_model.slug})
        response = self.CLIENT.get(po_create_url)

        # purchase order create page is rendered
        self.assertEqual(response.status_code, 200, msg="Fail to GET PO create page")

        # after successfully create a PO, redirect to list
        po_title = f'PO-Create-{randint(1000, 9999)}'
        redirect_response = self.CLIENT.post(po_create_url, data={'po_title': po_title}, follow=False)
        self.assertEqual(redirect_response.status_code, 302, msg="Create PO failed.")

        # check the redirect URL is correct
        redirect_url = urlparse(redirect_response.url)
        redirect_path = redirect_url.path
        list_path = reverse('django_ledger:po-list', kwargs={'entity_slug': entity_model.slug})
        self.assertEqual(redirect_path, list_path, msg="Fail to redirect properly after create a PO")

        # check the created PO is in the list
        response = self.CLIENT.get(redirect_response.url)
        self.assertContains(response, po_title)

    def test_purchase_order_detail(self):
        """
        Check elements on PO detail page.
        """
        self.login_client()
        entity_model: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)
        po_status_dict = dict(PurchaseOrderModel.PO_STATUS)

        po_model_qs = self.get_purchase_orders(entity_model)
        # check links, do it on every test PO
        for po_model in po_model_qs:
            po_detail_url = reverse('django_ledger:po-detail', kwargs={
                'entity_slug': entity_model.slug,
                'po_pk': po_model.uuid
            })
            po_update_url = reverse('django_ledger:po-update', kwargs={
                'entity_slug': entity_model.slug,
                'po_pk': po_model.uuid
            })
            po_delete_url = reverse('django_ledger:po-delete', kwargs={
                'entity_slug': entity_model.slug,
                'po_pk': po_model.uuid
            })
            status = po_model.po_status

            with self.assertNumQueries(5):
                response = self.CLIENT.get(po_detail_url)
            self.assertEqual(response.status_code, 200, msg=f"Error browsing PO {po_model.uuid} detail page.")

            # the correct status is displayed
            self.assertContains(response, po_status_dict[status])
            # update link is displayed
            self.assertContains(response, po_update_url)

            if status == po_model.PO_STATUS_DRAFT:
                self.assertContains(response, po_model.get_mark_as_review_html_id())
                # self.assertContains(response, po_delete_url)
                self.assertContains(response, po_model.get_mark_as_canceled_html_id())

            elif status == po_model.PO_STATUS_REVIEW:
                self.assertContains(response, po_model.get_mark_as_draft_html_id())
                self.assertContains(response, po_model.get_mark_as_approved_html_id())
                # self.assertContains(response, po_delete_url)
                self.assertContains(response, po_model.get_mark_as_canceled_html_id())

            elif status == po_model.PO_STATUS_APPROVED:
                self.assertContains(response, po_model.get_mark_as_fulfilled_html_id())
                self.assertContains(response, po_model.get_mark_as_void_html_id())

            # elif status == po_model.PO_STATUS_VOID:
            #     pass

            # check update page
            # with self.assertNumQueries(23):                 # WHY IS IT SO MANY??
            #     response = self.CLIENT.get(po_update_url)
            # self.assertEqual(response.status_code, 200, msg=f"Error browsing PO {po_model.uuid} update page.")

            # after successful update, redirect to detail page



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

A Journal Entry (JE) is the foundation of all double entry accounting and financial data of any EntityModel.
A JE encapsulates a collection of TransactionModel, which must contain two transactions at a minimum. Each transaction
must perform a DEBIT or a CREDIT to an AccountModel. The JE Model performs additional validation to make sure that
the sum of all DEBITs and the sum of all CREDITs are equal to keep the books balanced.

A JE by default will be un-posted, which means that simply creating a JE will have no effect on the EntityModel
books. This behavior allows for constant refinement and persistence of JEs in the database without any impact on the
books. Only Journal Entries contained within a *POSTED* LedgerModel (see LedgerModel for documentation) will have an
impact in the EntityModel finances.

The JournalEntryModel also carries an optional EntityUnitModel, which are logical user-defined labels which help
segregate the different financial statements into different business operations (see EntityUnitModel for documentation).
Examples of EntityModelUnits are offices, departments, divisions, etc. *The user may request financial statements by
unit*.

All JEs automatically generate a sequential Journal Entry Number, which takes into consideration the Fiscal Year of the
JournalEntryModel instance. This functionality enables a human-readable tracking mechanism which helps with audits. It
is also searchable and indexed to support quick searches and queries.

The JournalEntryModel is also responsible for validating the Financial Activity involved in the operations of the
business. Whenever an account with ASSET_CA_CASH role is involved in a Journal Entry (see roles for more details), the
JE is responsible for programmatically determine the kind of operation for the JE (Operating, Financing, Investing).
"""
import warnings
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from itertools import chain
from typing import Set, Union, Optional, Dict, Tuple, List
from uuid import uuid4, UUID

from django.core.exceptions import FieldError, ObjectDoesNotExist, ValidationError
from django.db import models, transaction, IntegrityError
from django.db.models import Q, Sum, QuerySet, F, Manager, Count
from django.db.models.functions import Coalesce
from django.db.models.signals import pre_save
from django.urls import reverse
from django.utils.timezone import localtime
from django.utils.translation import gettext_lazy as _

from django_ledger.io import roles
from django_ledger.io.io_core import get_localtime
from django_ledger.io.roles import (
    ASSET_CA_CASH, GROUP_CFS_FIN_DIVIDENDS, GROUP_CFS_FIN_ISSUING_EQUITY,
    GROUP_CFS_FIN_LT_DEBT_PAYMENTS, GROUP_CFS_FIN_ST_DEBT_PAYMENTS,
    GROUP_CFS_INVESTING_AND_FINANCING, GROUP_CFS_INVESTING_PPE,
    GROUP_CFS_INVESTING_SECURITIES,
    validate_roles
)
from django_ledger.models.accounts import CREDIT, DEBIT
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.entity import EntityStateModel, EntityModel
from django_ledger.models.ledger import LedgerModel
from django_ledger.models.mixins import CreateUpdateMixIn
from django_ledger.models.signals import (
    journal_entry_unlocked,
    journal_entry_locked,
    journal_entry_posted,
    journal_entry_unposted
)
from django_ledger.models.transactions import TransactionModelQuerySet
from django_ledger.settings import (
    DJANGO_LEDGER_JE_NUMBER_PREFIX,
    DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING,
    DJANGO_LEDGER_JE_NUMBER_NO_UNIT_PREFIX, DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR
)


class JournalEntryValidationError(ValidationError):
    pass


class JournalEntryModelQuerySet(QuerySet):
    """
    A custom QuerySet for working with Journal Entry models, providing additional
    convenience methods and validations for specific use cases.

    This class enhances Django's default QuerySet by adding tailored methods
    to manage and filter Journal Entries, such as handling posted, unposted,
    locked entries, and querying entries associated with specific ledgers.
    """

    def for_user(self, user_model) -> 'JournalEntryModelQuerySet':
        """
        Filters the JournalEntryModel queryset for the given user.

        - Superusers will have access to all journal entries.
        - Other authenticated users will only see entries for entities where
          they are admins or managers.

        Parameters
        ----------
        user_model : UserModel
            An authenticated Django user object.

        Returns
        -------
        JournalEntryModelQuerySet
            A filtered queryset restricted by the user's entity relationships.
        """
        if user_model.is_superuser:
            return self

        return self.filter(
            Q(ledger__entity__admin=user_model) |  # Entries for entities where the user is admin
            Q(ledger__entity__managers__in=[user_model])  # Entries for entities where the user is a manager
        )

    def create(self, verify_on_save: bool = False, force_create: bool = False, **kwargs) -> 'JournalEntryModelQuerySet':
        """
        Creates a new Journal Entry while enforcing business logic validations.

        This method overrides Django's default `create()` to ensure that Journal Entries
        cannot be created in a "posted" state unless explicitly overridden.
        Additionally, it offers optional pre-save verification.

        Parameters
        ----------
        verify_on_save : bool
            If True, performs a verification step before saving. This avoids
            additional database queries for validation when creating new entries.
            Should be used when the Journal Entry needs no transactional validation.
        force_create : bool
            If True, allows the creation of a Journal Entry even in a "posted"
            state. Use with caution and only if you are certain of the consequences.
        **kwargs : dict
            Additional keyword arguments passed to instantiate the Journal Entry model.

        Returns
        -------
        JournalEntryModel
            The newly created Journal Entry.

        Raises
        ------
        FieldError
            Raised if attempting to create a "posted" Journal Entry without
            setting `force_create=True`.
        """
        is_posted = kwargs.get('posted')
        if is_posted and not force_create:
            raise FieldError("Cannot create Journal Entries in a posted state without 'force_create=True'.")

        obj = self.model(**kwargs)
        self._for_write = True

        # Save the object with optional pre-save verification.
        obj.save(force_insert=True, using=self.db, verify=verify_on_save)
        return obj

    def posted(self) -> 'JournalEntryModelQuerySet':
        """
        Filters the QuerySet to include only "posted" Journal Entries.

        Returns
        -------
        JournalEntryModelQuerySet
            A filtered QuerySet containing only posted Journal Entries.
        """
        return self.filter(posted=True)

    def unposted(self) -> 'JournalEntryModelQuerySet':
        """
        Filters the QuerySet to include only "unposted" Journal Entries.

        Returns
        -------
        JournalEntryModelQuerySet
            A filtered QuerySet containing only unposted Journal Entries.
        """
        return self.filter(posted=False)

    def locked(self) -> 'JournalEntryModelQuerySet':
        """
        Filters the QuerySet to include only "locked" Journal Entries.

        Returns
        -------
        JournalEntryModelQuerySet
            A filtered QuerySet containing only locked Journal Entries.
        """
        return self.filter(locked=True)

    def unlocked(self) -> 'JournalEntryModelQuerySet':
        """
        Filters the QuerySet to include only "unlocked" Journal Entries.

        Returns
        -------
        JournalEntryModelQuerySet
            A filtered QuerySet containing only unlocked Journal Entries.
        """
        return self.filter(locked=False)

    def for_ledger(self, ledger_pk: Union[str, UUID, LedgerModel]) -> 'JournalEntryModelQuerySet':
        """
        Filters the QuerySet to include Journal Entries associated with a specific Ledger.

        Parameters
        ----------
        ledger_pk : str, UUID, or LedgerModel
            The LedgerModel instance, its UUID, or a string representation of the UUID
            to identify the Ledger.

        Returns
        -------
        JournalEntryModelQuerySet
            A filtered QuerySet of Journal Entries associated with the specified Ledger.
        """
        if isinstance(ledger_pk, LedgerModel):
            return self.filter(ledger=ledger_pk)
        return self.filter(ledger__uuid__exact=ledger_pk)


class JournalEntryModelManager(Manager):
    """
    A custom manager for the JournalEntryModel that extends Django's default
    Manager with additional query features. It allows complex query handling
    based on relationships to the `EntityModel` and the authenticated `UserModel`.

    This manager provides utility methods for generating filtered querysets
    (e.g., entries associated with specific users or entities), as well as
    annotations for convenience in query results.
    """

    def get_queryset(self) -> JournalEntryModelQuerySet:
        """
        Returns the default queryset for JournalEntryModel with additional
        annotations applied.

        Annotations:
        - `_entity_slug`: The slug of the related `EntityModel`.
        - `txs_count`: The count of transactions (related `TransactionModel` instances)
          for each journal entry.

        Returns
        -------
        JournalEntryModelQuerySet
            A custom queryset enhanced with annotations.
        """
        qs = JournalEntryModelQuerySet(self.model, using=self._db)
        return qs.annotate(
            _entity_uuid=F('ledger__entity_id'),
            _entity_slug=F('ledger__entity__slug'),  # Annotates the entity slug
            _entity_last_closing_date=F('ledger__entity__last_closing_date'),
            _ledger_is_locked=F('ledger__locked'),
            txs_count=Count('transactionmodel')  # Annotates the count of transactions
        )

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: EntityModel | str | UUID = None, **kwargs) -> JournalEntryModelQuerySet:
        """
        Filters the JournalEntryModel queryset for a specific entity and user.

        This method provides a way to fetch journal entries related to a specific
        `EntityModel`, identified by its slug or model instance, with additional
        filtering scoped to the user.

        Parameters
        ----------
        entity_model : str or EntityModel
            The slug of the entity (or an instance of `EntityModel`) used for filtering.
        Returns
        -------
        JournalEntryModelQuerySet
            A customized queryset containing journal entries associated with the
            given entity and restricted by the user's access permissions.
        """
        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(ledger__entity=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(ledger__entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(ledger__entity_id=entity_model)
        else:
            raise JournalEntryValidationError(
                message='Must provide EntityModel, slug or UUID',
            )
        return qs


class ActivityEnum(Enum):
    """
    Represents the database prefixes used for different types of accounting activities.

    Attributes
    ----------
    OPERATING : str
        Prefix for a Journal Entry categorized as an Operating Activity.
    INVESTING : str
        Prefix for a Journal Entry categorized as an Investing Activity.
    FINANCING : str
        Prefix for a Journal Entry categorized as a Financing Activity.
    """
    OPERATING = 'op'
    INVESTING = 'inv'
    FINANCING = 'fin'


class JournalEntryModelAbstract(CreateUpdateMixIn):
    """
    Abstract base model for handling journal entries in the bookkeeping system.

    Attributes
    ----------
    uuid : UUID
        A unique identifier (primary key) for the journal entry, generated using uuid4().
    je_number : str
        A human-readable, unique, alphanumeric identifier for the journal entry (e.g., Voucher or Document Number).
        Includes the fiscal year as a prefix for organizational purposes.
    timestamp : datetime
        The date of the journal entry, used for financial statements. This timestamp applies to associated transactions.
    description : str
        An optional user-defined description for the journal entry.
    entity_unit : EntityUnitModel
        A reference to a logical and self-contained structure within the `EntityModel`.
        Provides context for the journal entry. See `EntityUnitModel` documentation for details.
    activity : str
        Indicates the nature of the activity associated with the journal entry.
        Must be one of the predefined `ACTIVITIES` (e.g., Operating, Financing, Investing) and is programmatically determined.
    origin : str
        Describes the origin or trigger for the journal entry (e.g., reconciliations, migrations, auto-generated).
        Max length: 30 characters.
    posted : bool
        Determines whether the journal entry has been posted (affecting the books). Defaults to `False`.
    locked : bool
        Indicates whether the journal entry is locked, preventing further modifications. Defaults to `False`.
    ledger : LedgerModel
        A reference to the LedgerModel associated with this journal entry. This field is mandatory.
    is_closing_entry : bool
        Indicates if the journal entry is a closing entry. Defaults to `False`.
    """

    # Constants for activity types
    OPERATING_ACTIVITY = ActivityEnum.OPERATING.value
    FINANCING_OTHER = ActivityEnum.FINANCING.value
    INVESTING_OTHER = ActivityEnum.INVESTING.value
    INVESTING_SECURITIES = f'{ActivityEnum.INVESTING.value}_securities'
    INVESTING_PPE = f'{ActivityEnum.INVESTING.value}_ppe'
    FINANCING_STD = f'{ActivityEnum.FINANCING.value}_std'
    FINANCING_LTD = f'{ActivityEnum.FINANCING.value}_ltd'
    FINANCING_EQUITY = f'{ActivityEnum.FINANCING.value}_equity'
    FINANCING_DIVIDENDS = f'{ActivityEnum.FINANCING.value}_dividends'

    # Activity categories for dropdown
    ACTIVITIES = [
        (_('Operating'), (
            (OPERATING_ACTIVITY, _('Operating')),
        )),
        (_('Investing'), (
            (INVESTING_PPE, _('Purchase/Disposition of PPE')),
            (INVESTING_SECURITIES, _('Purchase/Disposition of Securities')),
            (INVESTING_OTHER, _('Investing Activity Other')),
        )),
        (_('Financing'), (
            (FINANCING_STD, _('Payoff of Short Term Debt')),
            (FINANCING_LTD, _('Payoff of Long Term Debt')),
            (FINANCING_EQUITY, _('Issuance of Common Stock, Preferred Stock or Capital Contribution')),
            (FINANCING_DIVIDENDS, _('Dividends or Distributions to Shareholders')),
            (FINANCING_OTHER, _('Financing Activity Other')),
        )),
    ]

    # Utility mappings for activity validation
    VALID_ACTIVITIES = list(chain.from_iterable([[a[0] for a in cat[1]] for cat in ACTIVITIES]))
    MAP_ACTIVITIES = dict(chain.from_iterable([[(a[0], cat[0]) for a in cat[1]] for cat in ACTIVITIES]))
    NON_OPERATIONAL_ACTIVITIES = [a for a in VALID_ACTIVITIES if ActivityEnum.OPERATING.value not in a]

    # Field definitions
    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    je_number = models.SlugField(max_length=25, editable=False, verbose_name=_('Journal Entry Number'))
    timestamp = models.DateTimeField(verbose_name=_('Timestamp'), default=localtime)
    description = models.CharField(max_length=70, blank=True, null=True, verbose_name=_('Description'))
    entity_unit = models.ForeignKey(
        'django_ledger.EntityUnitModel',
        on_delete=models.RESTRICT,
        blank=True,
        null=True,
        verbose_name=_('Associated Entity Unit')
    )
    activity = models.CharField(
        choices=ACTIVITIES,
        max_length=20,
        null=True,
        blank=True,
        editable=False,
        verbose_name=_('Activity')
    )
    origin = models.CharField(max_length=30, blank=True, null=True, verbose_name=_('Origin'))
    posted = models.BooleanField(default=False, verbose_name=_('Posted'))
    locked = models.BooleanField(default=False, verbose_name=_('Locked'))
    is_closing_entry = models.BooleanField(default=False)
    ledger = models.ForeignKey(
        'django_ledger.LedgerModel',
        verbose_name=_('Ledger'),
        related_name='journal_entries',
        on_delete=models.CASCADE
    )

    # Custom manager
    objects = JournalEntryModelManager.from_queryset(queryset_class=JournalEntryModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('Journal Entry')
        verbose_name_plural = _('Journal Entries')
        indexes = [
            models.Index(fields=['ledger']),
            models.Index(fields=['timestamp']),
            models.Index(fields=['activity']),
            models.Index(fields=['entity_unit']),
            models.Index(fields=['locked']),
            models.Index(fields=['posted']),
            models.Index(fields=['je_number']),
            models.Index(fields=['is_closing_entry']),
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._verified = False

    def __str__(self):
        if self.je_number:
            return f"JE: {self.je_number} (posted={self.posted}, locked={self.locked}) - Desc: {self.description or ''}"
        return f"JE ID: {self.pk} (posted={self.posted}, locked={self.locked}) - Desc: {self.description or ''}"

    @property
    def entity_uuid(self):
        try:
            return getattr(self, '_entity_uuid')
        except AttributeError:
            pass
        return self.ledger.entity_id

    @property
    def entity_slug(self):
        """
        Retrieves the unique slug associated with the entity.

        The property first attempts to return the value stored in the `_entity_slug`
        attribute if it exists. If `_entity_slug` is not set, it falls back to the
        `ledger.entity.slug` attribute.

        Returns:
            str: The slug value from `_entity_slug` if available, or `ledger.entity.slug` otherwise.
        """
        try:
            return getattr(self, '_entity_slug')
        except AttributeError:
            pass
        return self.ledger.entity.slug

    @property
    def entity_model(self) -> EntityModel:
        """
        Provides access to the `EntityModel` related to the JournalEntryModel.

        Returns
        -------
        EntityModel
            The `EntityModel` instance linked to the instance LedgerModel.
        """
        return self.ledger.entity

    @property
    def entity_last_closing_date(self) -> Optional[date]:
        """
            Retrieves the last closing date for an entity, if available.

            This property returns the date of the most recent closing event
            associated with the entity. If no closing date exists, the
            result will be None.

            Returns
            -------
            Optional[date]
                The date of the last closing event, or None if no closing
                date is available.
        """
        return self.get_entity_last_closing_date()

    def validate_for_entity(self, entity_model: Union[EntityModel, str, UUID], raise_exception: bool = True) -> bool:
        """
        Validates whether the given entity_model owns thr Journal Entry instance.

        This method checks if the provided entity_model owns the Journal Entry model instance.
        The entity_model can be of type `EntityModel`, `str`, or
        `UUID`. The method performs type-specific checks to ensure proper validation
        and returns the validation result.

        Parameters
        ----------
        entity_model : Union[EntityModel, str, UUID]
            The entity to validate against. It can either be an instance of the
            `EntityModel`, a string representation of a UUID, or a UUID object.

        Returns
        -------
        bool
            A boolean value. True if the given entity_model corresponds to the current
            entity's UUID, otherwise False.
        """
        if isinstance(entity_model, str):
            is_valid = str(self.entity_uuid) == entity_model
        elif isinstance(entity_model, UUID):
            is_valid = self.entity_uuid == entity_model
        else:
            is_valid = self.entity_uuid == entity_model.uuid

        if not is_valid and raise_exception:
            raise JournalEntryValidationError(
                message='The Journal Entry does not belong to the provided entity.'
            )
        return is_valid

    def ledger_is_locked(self):
        """
        Determines whether the ledger is locked.

        This method checks the current state of the ledger to determine if it is
        locked and unavailable for further operations. It looks for an annotated
        attribute `_ledger_is_locked` and returns its value if found. If the
        attribute is not set, it delegates the check to the actual `is_locked`
        method of the `ledger` object.

        Returns
        -------
        bool
            A boolean value indicating whether the ledger is locked.
        """
        try:
            return getattr(self, '_ledger_is_locked')
        except AttributeError:
            pass
        return self.ledger.is_locked()

    def can_post(self, ignore_verify: bool = True) -> bool:
        """Determines if the journal entry can be posted."""
        return all([
            self.is_locked(),
            not self.is_posted(),
            self.is_verified() if not ignore_verify else True,  # avoids db queries, will be verified before saving
            not self.ledger_is_locked(),
            not self.is_in_locked_period()
        ])

    def can_unpost(self) -> bool:
        """Checks if the journal entry can be un-posted."""
        return all([
            self.is_posted(),
            not self.ledger_is_locked(),
            not self.is_in_locked_period()
        ])

    def can_lock(self) -> bool:
        """Determines if the journal entry can be locked."""
        return all([
            not self.is_locked(),
            not self.ledger_is_locked()
        ])

    def can_unlock(self) -> bool:
        """Checks if the journal entry can be unlocked."""
        return all([
            self.is_locked(),
            not self.is_posted(),
            not self.is_in_locked_period(),
            not self.ledger_is_locked()
        ])

    def can_delete(self) -> bool:
        """Checks if the journal entry can be deleted."""
        return all([
            not self.is_locked(),
            not self.is_posted(),
        ])

    def can_edit(self) -> bool:
        """Checks if the journal entry is editable."""
        return all([
            not self.is_locked(),

        ])

    def is_posted(self) -> bool:
        """Returns whether the journal entry has been posted."""
        return self.posted is True

    def is_in_locked_period(self, new_timestamp: Optional[Union[date, datetime]] = None) -> bool:
        """
        Checks if the current Journal Entry falls within a locked period.

        Parameters
        ----------
            new_timestamp: Optional[Union[date, datetime]])
                An optional date or timestamp to be checked instead of the current timestamp.

        Returns
        -------
            bool: True if the Journal Entry is in a locked period, otherwise False.
        """
        last_closing_date = self.entity_last_closing_date
        if last_closing_date is not None:
            if not new_timestamp:
                return last_closing_date >= self.timestamp.date()
            elif isinstance(new_timestamp, datetime):
                return last_closing_date >= new_timestamp.date()
            return last_closing_date >= new_timestamp
        return False

    def is_locked(self) -> bool:
        """
        Determines if the Journal Entry is locked.

        A Journal Entry is considered locked if it is posted, explicitly marked
        as locked, falls within a locked period, or the associated ledger is locked.

        Returns:
            bool: True if the Journal Entry is locked, otherwise False.
        """
        return self.is_posted() or any([
            self.locked,
            self.is_in_locked_period(),
            self.ledger_is_locked()
        ])

    def is_verified(self) -> bool:
        """
        Checks if the Journal Entry is verified.

        Returns:
            bool: True if the Journal Entry is verified, otherwise False.
        """
        return self._verified

    def is_balance_valid(self, txs_qs: TransactionModelQuerySet, raise_exception: bool = True) -> bool:
        """
        Validates whether the DEBITs and CREDITs of the transactions balance correctly.

        Parameters:
            txs_qs (TransactionModelQuerySet): A QuerySet containing transactions to validate.
            raise_exception (bool): Whether to raise a JournalEntryValidationError if the validation fails.

        Returns:
            bool: True if the transactions are balanced, otherwise False.

        Raises:
            JournalEntryValidationError: If the transactions are not balanced and raise_exception is True.
        """
        if len(txs_qs) > 0:
            balances = self.get_txs_balances(txs_qs=txs_qs, as_dict=True)
            is_valid = balances[CREDIT] == balances[DEBIT]
            if not is_valid:
                if raise_exception:
                    raise JournalEntryValidationError(
                        message='Balance of {0} CREDITs are {1} does not match DEBITs {2}.'.format(
                            self,
                            balances[CREDIT],
                            balances[DEBIT]
                        )
                    )
            return is_valid
        return True

    def is_txs_qs_coa_valid(self, txs_qs: TransactionModelQuerySet, raise_exception: bool = True) -> bool:
        """
        Validates that all transactions in the QuerySet are associated with the same Chart of Accounts (COA).

        Parameters:
            txs_qs (TransactionModelQuerySet): A QuerySet containing transactions to validate.

        Returns:
            bool: True if all transactions have the same Chart of Accounts, otherwise False.
        """
        if len(txs_qs) > 0:
            coa_count = len(set(tx.coa_id for tx in txs_qs))
            is_valid = coa_count == 1
            if not is_valid and raise_exception:
                raise JournalEntryValidationError(
                    message='All transactions in the QuerySet must be associated with the same Chart of Accounts.'
                )
            return is_valid
        return True

    def is_txs_qs_valid(self, txs_qs: TransactionModelQuerySet, raise_exception: bool = True) -> bool:
        """
        Validates whether the given Transaction QuerySet belongs to the current Journal Entry.

        Parameters:
            txs_qs (TransactionModelQuerySet): A QuerySet containing transactions to validate.
            raise_exception (bool): Whether to raise a JournalEntryValidationError if the validation fails.

        Returns:
            bool: True if all transactions belong to the Journal Entry, otherwise False.

        Raises:
            JournalEntryValidationError: If validation fails and raise_exception is True.
        """
        if not isinstance(txs_qs, TransactionModelQuerySet):
            raise JournalEntryValidationError('Must pass an instance of TransactionModelQuerySet')

        is_valid = all(tx.journal_entry_id == self.uuid for tx in txs_qs)
        if not is_valid and raise_exception:
            raise JournalEntryValidationError(
                f'Invalid TransactionModelQuerySet. All transactions must be associated with Journal Entry {self.uuid}.'
            )
        return is_valid

    def is_cash_involved(self, txs_qs: Optional[TransactionModelQuerySet] = None) -> bool:
        """
        Checks if the transactions involve cash assets.

        Parameters:
            txs_qs (Optional[TransactionModelQuerySet]): Transactions to evaluate. If None, defaults to class behavior.

        Returns:
            bool: True if cash assets are involved, otherwise False.
        """
        return roles.ASSET_CA_CASH in self.get_txs_roles(txs_qs)

    def is_operating(self) -> bool:
        """
        Checks if the Journal Entry is categorized as an operating activity.

        Returns:
            bool: True if the activity is operating, otherwise False.
        """
        return self.activity in [self.OPERATING_ACTIVITY]

    def is_financing(self) -> bool:
        """
        Checks if the Journal Entry is categorized as a financing activity.

        Returns:
            bool: True if the activity is financing, otherwise False.
        """
        return self.activity in [
            self.FINANCING_EQUITY,
            self.FINANCING_LTD,
            self.FINANCING_DIVIDENDS,
            self.FINANCING_STD,
            self.FINANCING_OTHER
        ]

    def is_investing(self) -> bool:
        """
        Checks if the Journal Entry is categorized as an investing activity.

        Returns:
            bool: True if the activity is investing, otherwise False.
        """
        return self.activity in [
            self.INVESTING_SECURITIES,
            self.INVESTING_PPE,
            self.INVESTING_OTHER
        ]

    def get_entity_unit_name(self, no_unit_name: str = "") -> str:
        """
        Retrieves the name of the entity unit associated with the Journal Entry.

        Parameters:
            no_unit_name (str): The fallback name to return if no unit is associated.

        Returns:
            str: The name of the entity unit, or the fallback provided.
        """
        if self.entity_unit_id:
            return self.entity_unit.name
        return no_unit_name

    def get_entity_last_closing_date(self) -> Optional[date]:
        """
        Retrieves the last closing date for the entity associated with the Journal Entry.

        Returns:
            Optional[date]: The last closing date if one exists, otherwise None.
        """
        try:
            return getattr(self, '_entity_last_closing_date')
        except AttributeError:
            pass
        return self.ledger.entity.last_closing_date

    def mark_as_posted(self,
                       commit: bool = False,
                       verify: bool = True,
                       force_lock: bool = False,
                       raise_exception: bool = False,
                       **kwargs):
        """
        Posted transactions show on the EntityModel ledger and financial statements.
        Parameters
        ----------
        commit: bool
            Commits changes into the Database, Defaults to False.
        verify: bool
            Verifies JournalEntryModel before marking as posted. Defaults to False.
        force_lock: bool
            Forces to lock the JournalEntry before is posted.
        raise_exception: bool
            Raises JournalEntryValidationError if cannot post. Defaults to False.
        kwargs: dict
            Additional keyword arguments.
        """
        if verify and not self.is_verified():
            txs_qs, verified = self.verify()
            if not len(txs_qs):
                if raise_exception:
                    raise JournalEntryValidationError(
                        message=_('Cannot post an empty Journal Entry.')
                    )
                return
        if force_lock and not self.is_locked():
            try:
                self.mark_as_locked(commit=False, raise_exception=True)
            except JournalEntryValidationError as e:
                if raise_exception:
                    raise e
                return
        if not self.can_post(ignore_verify=False):
            if raise_exception:
                raise JournalEntryValidationError(f'Journal Entry {self.uuid} cannot post.'
                                                  f' Is verified: {self.is_verified()}')
            return
        if not self.is_posted():
            self.posted = True
            if self.is_posted():
                if commit:
                    self.save(verify=False,
                              update_fields=[
                                  'posted',
                                  'locked',
                                  'activity',
                                  'updated'
                              ])
            journal_entry_posted.send_robust(sender=self.__class__,
                                             instance=self,
                                             commited=commit,
                                             **kwargs)

    def post(self, **kwargs):
        """
        Proxy function for `mark_as_posted` method.
        """
        return self.mark_as_posted(**kwargs)

    def mark_as_unposted(self, commit: bool = False, raise_exception: bool = False, **kwargs):
        """
        Un-posted JournalEntryModels do not show on the EntityModel ledger and financial statements.

        Parameters
        ----------
        commit: bool
            Commits changes into the Database, Defaults to False.
        raise_exception: bool
            Raises JournalEntryValidationError if cannot post. Defaults to False.
        kwargs: dict
            Additional keyword arguments.
        """
        if not self.can_unpost():
            if raise_exception:
                raise JournalEntryValidationError(f'Journal Entry {self.uuid} cannot unpost.')
            return
        if self.is_posted():
            self.posted = False
            self.activity = None
            if not self.is_posted():
                if commit:
                    self.save(
                        verify=False,
                        update_fields=[
                            'posted',
                            'activity',
                            'updated'
                        ]
                    )
            journal_entry_unposted.send_robust(sender=self.__class__,
                                               instance=self,
                                               commited=commit,
                                               **kwargs)

    def unpost(self, **kwargs):
        """
        Proxy function for `mark_as_unposted` method.
        """
        return self.mark_as_unposted(**kwargs)

    def mark_as_locked(self, commit: bool = False, raise_exception: bool = False, **kwargs):
        """
        Locked JournalEntryModels do not allow transactions to be edited.

        Parameters
        ----------
        commit: bool
            Commits changes into the Database, Defaults to False.
        raise_exception: bool
            Raises JournalEntryValidationError if cannot lock. Defaults to False.
        kwargs: dict
            Additional keyword arguments.
        """
        if not self.can_lock():
            if raise_exception:
                if raise_exception:
                    raise JournalEntryValidationError(f'Journal Entry {self.uuid} is already locked.')
                return
        if not self.is_locked():
            self.generate_activity(force_update=True)
            self.locked = True
            if self.is_locked():
                if commit:
                    self.save(verify=False)
                journal_entry_locked.send_robust(sender=self.__class__,
                                                 instance=self,
                                                 commited=commit,
                                                 **kwargs)

    def lock(self, **kwargs):
        """
        Proxy function for `mark_as_locked` method.
        """
        return self.mark_as_locked(**kwargs)

    def mark_as_unlocked(self, commit: bool = False, raise_exception: bool = False, **kwargs):
        """
        Unlocked JournalEntryModels allow transactions to be edited.

        Parameters
        ----------
        commit: bool
            Commits changes into the Database, Defaults to False.
        raise_exception: bool
            Raises JournalEntryValidationError if cannot lock. Defaults to False.
        kwargs: dict
            Additional keyword arguments.
        """
        if not self.can_unlock():
            if raise_exception:
                raise JournalEntryValidationError(f'Journal Entry {self.uuid} is already unlocked.')
            return
        if self.is_locked():
            self.locked = False
            self.activity = None
            if not self.is_locked():
                if commit:
                    self.save(verify=False)
                journal_entry_unlocked.send_robust(sender=self.__class__,
                                                   instance=self,
                                                   commited=commit,
                                                   **kwargs)

    def unlock(self, **kwargs):
        """
        Proxy function for `mark_as_unlocked` method.
        """
        return self.mark_as_unlocked(**kwargs)

    def get_transaction_queryset(self, select_accounts: bool = True) -> TransactionModelQuerySet:
        """
        Retrieves the `TransactionModelQuerySet` associated with this `JournalEntryModel` instance.

        Parameters
        ----------
        select_accounts : bool, optional
            If True, prefetches the related `AccountModel` for each transaction. Defaults to True.

        Returns
        -------
        TransactionModelQuerySet
            A queryset containing transactions related to this journal entry. If `select_accounts` is
            True, the accounts are included in the query as well.
        """
        if select_accounts:
            return self.transactionmodel_set.all().select_related('account')
        return self.transactionmodel_set.all()

    def get_txs_balances(
            self,
            txs_qs: Optional[TransactionModelQuerySet] = None,
            as_dict: bool = False
    ) -> Union[TransactionModelQuerySet, Dict[str, Decimal]]:
        """
        Calculates the total CREDIT and DEBIT balances for the journal entry.

        This method performs an aggregate database query to compute the sum of CREDITs and
        DEBITs across the transactions related to this journal entry. Optionally, a pre-fetched
        `TransactionModelQuerySet` can be supplied for efficiency. Validation is performed to
        ensure that all transactions belong to this journal entry.

        Parameters
        ----------
        txs_qs : TransactionModelQuerySet, optional
            A pre-fetched queryset of transactions. If None, the queryset is fetched automatically.
        as_dict : bool, optional
            If True, returns the results as a dictionary with keys "credit" and "debit". Defaults to False.

        Returns
        -------
        Union[TransactionModelQuerySet, Dict[str, Decimal]]
            If `as_dict` is False, returns a queryset of aggregated balances. If `as_dict` is True,
            returns a dictionary containing the CREDIT and DEBIT totals.

        Raises
        ------
        JournalEntryValidationError
            If the provided queryset is invalid or does not belong to this journal entry.
        """
        if not txs_qs:
            txs_qs = self.get_transaction_queryset(select_accounts=False)
        elif not isinstance(txs_qs, TransactionModelQuerySet):
            raise JournalEntryValidationError(
                f"Expected a TransactionModelQuerySet, got {type(txs_qs).__name__}"
            )
        elif not self.is_txs_qs_valid(txs_qs):
            raise JournalEntryValidationError(
                "Invalid TransactionModelQuerySet. All transactions must belong to the same journal entry."
            )

        balances = txs_qs.values('tx_type').annotate(
            amount__sum=Coalesce(Sum('amount'), Decimal('0.00'), output_field=models.DecimalField())
        )

        if as_dict:
            return {tx['tx_type']: tx['amount__sum'] for tx in balances}
        return balances

    def get_txs_roles(
            self,
            txs_qs: Optional[TransactionModelQuerySet] = None,
            exclude_cash_role: bool = False
    ) -> Set[str]:
        """
        Retrieves the set of account roles involved in the journal entry's transactions.

        This method extracts the roles associated with the accounts linked to each transaction.
        Optionally, the CASH role can be excluded from the results.

        Parameters
        ----------
        txs_qs : TransactionModelQuerySet, optional
            A pre-fetched queryset of transactions. If None, the queryset is fetched automatically.
        exclude_cash_role : bool, optional
            If True, excludes the CASH role from the result. Defaults to False.

        Returns
        -------
        Set[str]
            A set of account roles associated with this journal entry's transactions.
        """
        if not txs_qs:
            txs_qs = self.get_transaction_queryset(select_accounts=True)
        else:
            self.is_txs_qs_valid(txs_qs)

        roles = {tx.account.role for tx in txs_qs}

        if exclude_cash_role:
            roles.discard(ASSET_CA_CASH)

        return roles

    def has_activity(self) -> bool:
        """
        Checks if the journal entry has an associated activity.

        Returns
        -------
        bool
            True if an activity is defined for the journal entry, otherwise False.
        """
        return self.activity is not None

    def get_activity_name(self) -> Optional[str]:
        """
        Gets the name of the activity associated with this journal entry.

        The activity indicates its categorization based on GAAP (e.g., operating, investing, financing).

        Returns
        -------
        Optional[str]
            The activity name if defined, otherwise None.
        """
        if self.activity:
            if self.is_operating():
                return ActivityEnum.OPERATING.value
            elif self.is_investing():
                return ActivityEnum.INVESTING.value
            elif self.is_financing():
                return ActivityEnum.FINANCING.value
        return None

    @classmethod
    def get_activity_from_roles(
            cls,
            role_set: Union[List[str], Set[str]],
            validate: bool = False,
            raise_exception: bool = True
    ) -> Optional[str]:
        """
        Determines the financial activity type (e.g., operating, investing, financing)
        based on a set of account roles.

        Parameters
        ----------
        role_set : Union[List[str], Set[str]]
            The set of roles to analyze.
        validate : bool, optional
            If True, validates the roles before analysis. Defaults to False.
        raise_exception : bool, optional
            If True, raises an exception if multiple activities are detected. Defaults to True.

        Returns
        -------
        Optional[str]
            The detected activity name, or None if no activity type is matched.

        Raises
        ------
        JournalEntryValidationError
            If multiple activities are detected and `raise_exception` is True.
        """
        if validate:
            role_set = validate_roles(roles=role_set)
        else:
            if isinstance(role_set, list):
                role_set = set(role_set)

        activity = None

        # no roles involved
        if not len(role_set):
            return

        # determining if investing....
        is_investing_for_ppe = all([
            # all roles must be in group
            all([r in GROUP_CFS_INVESTING_PPE for r in role_set]),
            # at least one role
            sum([r in GROUP_CFS_INVESTING_PPE for r in role_set]) > 0,
            # at least one role
            # sum([r in GROUP_CFS_INV_LTD_OF_PPE for r in role_set]) > 0,
        ])
        is_investing_for_securities = all([
            # all roles must be in group
            all([r in GROUP_CFS_INVESTING_SECURITIES for r in role_set]),
            # at least one role
            sum([r in GROUP_CFS_INVESTING_SECURITIES for r in role_set]) > 0,
            # at least one role
            # sum([r in GROUP_CFS_INV_LTD_OF_SECURITIES for r in role_set]) > 0,
        ])

        # IS INVESTING OTHERS....?

        # determining if financing...
        is_financing_dividends = all([r in GROUP_CFS_FIN_DIVIDENDS for r in role_set])
        is_financing_issuing_equity = all([r in GROUP_CFS_FIN_ISSUING_EQUITY for r in role_set])
        is_financing_st_debt = all([r in GROUP_CFS_FIN_ST_DEBT_PAYMENTS for r in role_set])
        is_financing_lt_debt = all([r in GROUP_CFS_FIN_LT_DEBT_PAYMENTS for r in role_set])

        is_operating = all([r not in GROUP_CFS_INVESTING_AND_FINANCING for r in role_set])

        if sum([
            is_investing_for_ppe,
            is_investing_for_securities,
            is_financing_lt_debt,
            is_financing_st_debt,
            is_financing_issuing_equity,
            is_financing_dividends,
            is_operating
        ]) > 1:
            if raise_exception:
                raise JournalEntryValidationError(
                    f'Multiple activities detected in roles JE {role_set}.')
        else:
            if is_investing_for_ppe:
                activity = cls.INVESTING_PPE
            elif is_investing_for_securities:
                activity = cls.INVESTING_SECURITIES
            elif is_financing_st_debt:
                activity = cls.FINANCING_STD
            elif is_financing_lt_debt:
                activity = cls.FINANCING_LTD
            elif is_financing_issuing_equity:
                activity = cls.FINANCING_EQUITY
            elif is_financing_dividends:
                activity = cls.FINANCING_DIVIDENDS
            elif is_operating:
                activity = cls.OPERATING_ACTIVITY
            else:
                if raise_exception:
                    raise JournalEntryValidationError(f'No activity match for roles {role_set}.'
                                                      'Split into multiple Journal Entries or check'
                                                      ' your account selection.')
        return activity

    def generate_activity(self,
                          txs_qs: Optional[TransactionModelQuerySet] = None,
                          raise_exception: bool = True,
                          force_update: bool = False) -> Optional[str]:
        """
        Generates the activity for the Journal Entry model based on its transactions.

        Parameters
        ----------
        txs_qs : Optional[TransactionModelQuerySet], default None
            Queryset of TransactionModel instances for validation. If None, transactions are queried.
        raise_exception : bool, default True
            Determines whether exceptions are raised during processing.
        force_update : bool, default False
            Forces the regeneration of activity even if it exists.

        Returns
        -------
        Optional[str]
            Generated activity or None if not applicable.
        """
        if not self._state.adding:
            if raise_exception and self.is_closing_entry:
                raise_exception = False

            if any([
                not self.has_activity(),
                not self.is_locked(),
                force_update
            ]):

                txs_is_valid = True
                if not txs_qs:
                    txs_qs = self.get_transaction_queryset(select_accounts=False)
                else:
                    try:
                        txs_is_valid = self.is_txs_qs_valid(txs_qs=txs_qs, raise_exception=raise_exception)
                    except JournalEntryValidationError as e:
                        if raise_exception:
                            raise e

                if txs_is_valid:
                    cash_is_involved = self.is_cash_involved(txs_qs=txs_qs)
                    if not cash_is_involved:
                        self.activity = None
                    else:
                        role_list = self.get_txs_roles(txs_qs, exclude_cash_role=True)
                        self.activity = self.get_activity_from_roles(
                            role_set=role_list,
                            raise_exception=raise_exception
                        )
        return self.activity

    # todo: add entity_model as parameter on all functions...
    # todo: outsource this function to EntityStateModel...?...
    def _get_next_state_model(self, raise_exception: bool = True) -> Optional[EntityStateModel]:
        """
        Retrieves or creates the next state model for the Journal Entry.

        Parameters
        ----------
        raise_exception : bool, default True
            Determines if exceptions should be raised when the entity state is not found.

        Returns
        -------
        EntityStateModel
            The state model with an incremented sequence.
        """
        entity_model = self.entity_model
        fy_key = entity_model.get_fy_for_date(dt=self.timestamp)

        try:
            LOOKUP = {
                'entity_model_id__exact': self.entity_uuid,
                'entity_unit_id__exact': self.entity_unit_id,
                'fiscal_year': fy_key,
                'key__exact': EntityStateModel.KEY_JOURNAL_ENTRY
            }

            state_model_qs = EntityStateModel.objects.filter(**LOOKUP).select_related(
                'entity_model').select_for_update()
            state_model = state_model_qs.get()
            state_model.sequence = F('sequence') + 1
            state_model.save()
            state_model.refresh_from_db()
            return state_model

        except ObjectDoesNotExist:
            LOOKUP = {
                'entity_model_id': entity_model.uuid,
                'entity_unit_id': self.entity_unit_id,
                'fiscal_year': fy_key,
                'key': EntityStateModel.KEY_JOURNAL_ENTRY,
                'sequence': 1
            }
            state_model = EntityStateModel.objects.create(**LOOKUP)
            return state_model

        except IntegrityError as e:
            if raise_exception:
                raise e

    def can_generate_je_number(self) -> bool:
        """
        Checks if a Journal Entry Number can be generated.

        Returns
        -------
        bool
            True if the Journal Entry can generate a JE number, otherwise False.
        """
        return all([
            self.ledger_id,
            not self.je_number
        ])

    def generate_je_number(self, commit: bool = False) -> str:
        """
        Generates the Journal Entry number in an atomic transaction.

        Parameters
        ----------
        commit : bool, default False
            Saves the generated JE number in the database.

        Returns
        -------
        str
            The generated or existing JE number.
        """
        if self.can_generate_je_number():

            with transaction.atomic():

                state_model = None
                while not state_model:
                    state_model = self._get_next_state_model(raise_exception=False)

                if self.entity_unit_id:
                    unit_prefix = self.entity_unit.document_prefix
                else:
                    unit_prefix = DJANGO_LEDGER_JE_NUMBER_NO_UNIT_PREFIX

                seq = str(state_model.sequence).zfill(DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING)
                self.je_number = f'{DJANGO_LEDGER_JE_NUMBER_PREFIX}-{state_model.fiscal_year}-{unit_prefix}-{seq}'

                if commit:
                    self.save(update_fields=['je_number'])

        return self.je_number

    def verify(self,
               txs_qs: Optional[TransactionModelQuerySet] = None,
               force_verify: bool = False,
               raise_exception: bool = True,
               **kwargs) -> Tuple[TransactionModelQuerySet, bool]:

        """
        Verifies the validity of the Journal Entry model instance.

        Parameters
        ----------
        txs_qs : Optional[TransactionModelQuerySet], default None
            Queryset of TransactionModel instances to validate. If None, transactions are queried.
        force_verify : bool, default False
            Forces re-verification even if already verified.
        raise_exception : bool, default True
            Determines if exceptions are raised on validation failure.
        kwargs : dict
            Additional options.

        Returns
        -------
        Tuple[TransactionModelQuerySet, bool]
            - The TransactionModelQuerySet associated with the JournalEntryModel.
            - A boolean indicating whether verification was successful.
        """

        if not self.is_verified() or force_verify:
            self._verified = False

            # fetches JEModel TXS QuerySet if not provided....
            if not txs_qs:
                txs_qs = self.get_transaction_queryset()
                is_txs_qs_valid = True
            else:
                try:
                    # if provided, it is verified...
                    is_txs_qs_valid = self.is_txs_qs_valid(raise_exception=raise_exception, txs_qs=txs_qs)
                except JournalEntryValidationError as e:
                    raise e

            # CREDIT/DEBIT Balance validation...
            try:
                is_balance_valid = self.is_balance_valid(txs_qs=txs_qs)
                if not is_balance_valid:
                    raise JournalEntryValidationError('Transaction balances are not valid!')
            except JournalEntryValidationError as e:
                raise e

            # Transaction CoA if valid...

            try:
                is_coa_valid = self.is_txs_qs_coa_valid(txs_qs=txs_qs)
                if not is_coa_valid:
                    raise JournalEntryValidationError('Transaction COA is not valid!')
            except JournalEntryValidationError as e:
                raise e

            # if not len(txs_qs):
            #     if raise_exception:
            #         raise JournalEntryValidationError('Journal entry has no transactions.')

            # if len(txs_qs) < 2:
            #     if raise_exception:
            #         raise JournalEntryValidationError('At least two transactions required.')

            if all([
                is_balance_valid,
                is_txs_qs_valid,
                is_coa_valid
            ]):
                # activity flag...
                self.generate_activity(txs_qs=txs_qs, raise_exception=raise_exception)
                self._verified = True
                return txs_qs, self.is_verified()
        return self.get_transaction_queryset(), self.is_verified()

    def clean(self,
              verify: bool = False,
              raise_exception: bool = True,
              txs_qs: Optional[TransactionModelQuerySet] = None) -> Tuple[TransactionModelQuerySet, bool]:
        """
        Cleans the JournalEntryModel instance, optionally verifying it and generating a Journal Entry (JE) number if required.

        Parameters
        ----------
        verify : bool, optional
            If True, attempts to verify the JournalEntryModel during the cleaning process. Default is False.
        raise_exception : bool, optional
            If True, raises an exception when the instance fails verification. Default is True.
        txs_qs : TransactionModelQuerySet, optional
            A pre-fetched TransactionModelQuerySet. If provided, avoids additional database queries. The provided queryset is
            validated against the JournalEntryModel instance.

        Returns
        -------
        Tuple[TransactionModelQuerySet, bool]
            A tuple containing:
            - The validated TransactionModelQuerySet for the JournalEntryModel instance.
            - A boolean indicating whether the instance passed verification.

        Raises
        ------
        JournalEntryValidationError
            If the instance has a timestamp in the future and is posted, or if verification fails and `raise_exception` is True.
        """

        if txs_qs:
            self.is_txs_qs_valid(txs_qs=txs_qs)

        if not self.timestamp:
            self.timestamp = get_localtime()
        elif all([
            self.timestamp,
            self.timestamp > get_localtime(),
            self.is_posted()
        ]):
            raise JournalEntryValidationError(message='Cannot Post JE Models with timestamp in the future.')

        self.generate_je_number(commit=True)
        if verify:
            txs_qs, verified = self.verify()
            return txs_qs, self.is_verified()
        return self.get_transaction_queryset(), self.is_verified()

    def get_delete_message(self) -> str:
        """
        Generates a confirmation message for deleting the JournalEntryModel instance.

        Returns
        -------
        str
            A confirmation message including the Journal Entry number and Ledger name.
        """
        return _(f'Are you sure you want to delete JournalEntry Model {self.je_number} on Ledger {self.ledger.name}?')

    def delete(self, **kwargs):
        """
        Deletes the JournalEntryModel instance, ensuring it is allowed to be deleted.

        Parameters
        ----------
        **kwargs : dict
            Additional arguments passed to the parent delete method.

        Raises
        ------
        JournalEntryValidationError
            If the instance is not eligible for deletion.
        """
        if not self.can_delete():
            raise JournalEntryValidationError(
                message=_(f'JournalEntryModel {self.uuid} cannot be deleted...')
            )
        return super().delete(**kwargs)

    def save(
            self,
            verify: bool = True,
            post_on_verify: bool = False,
            *args,
            **kwargs
    ):
        """
        Saves the JournalEntryModel instance, with optional verification and posting prior to saving.

        Parameters
        ----------
        verify : bool, optional
            If True, verifies the transactions of the JournalEntryModel before saving. Default is True.
        post_on_verify : bool, optional
            If True, posts the JournalEntryModel if verification is successful and `can_post()` is True. Default is False.

        Returns
        -------
        JournalEntryModel
            The saved JournalEntryModel instance.

        Raises
        ------
        JournalEntryValidationError
            If the instance fails verification or encounters an issue during save.
        """
        try:
            # Generate the Journal Entry number prior to verification and saving
            self.generate_je_number(commit=False)

            if verify:
                txs_qs, is_verified = self.clean(verify=True)
                if is_verified and post_on_verify:
                    # Mark as posted if verification succeeds and posting is requested
                    self.mark_as_posted(commit=False, verify=False, force_lock=True, raise_exception=True)

        except ValidationError as e:
            if self.can_unpost():
                self.mark_as_unposted(raise_exception=True)
            raise JournalEntryValidationError(
                f'Error validating Journal Entry ID: {self.uuid}: {e.message}'
            )
        except Exception as e:
            # Safety net for unexpected errors during save
            self.posted = False
            self._verified = False
            self.save(update_fields=['posted', 'updated'], verify=False)
            raise JournalEntryValidationError(e)

        # Prevent saving an unverified Journal Entry
        if not self.is_verified() and verify:
            raise JournalEntryValidationError(message='Cannot save an unverified Journal Entry.')

        return super(JournalEntryModelAbstract, self).save(*args, **kwargs)

    # URLS Generation...

    def get_absolute_url(self) -> str:
        """
        Generates the URL to view the details of the journal entry.

        Returns
        -------
        str
            The absolute URL for the journal entry details.
        """
        return reverse('django_ledger:je-detail', kwargs={
            'je_pk': self.uuid,
            'ledger_pk': self.ledger_id,
            'entity_slug': self.entity_slug
        })

    def get_detail_url(self) -> str:
        """
        Generates the detail URL for the journal entry.

        Returns
        -------
        str
            The URL for updating or viewing journal entry details.
        """
        return self.get_absolute_url()

    def get_journal_entry_list_url(self) -> str:
        """
        Constructs the URL to access the list of journal entries
        associated with a specific ledger and entity.

        Returns
        -------
        str
            The URL for the journal entry list.
        """
        return reverse('django_ledger:je-list', kwargs={
            'entity_slug': self.entity_slug,
            'ledger_pk': self.ledger_id
        })

    def get_journal_entry_create_url(self) -> str:
        """
        Constructs the URL to create a new journal entry
        associated with a specific ledger and entity.

        Returns
        -------
        str
            The URL to create a journal entry.
        """
        return reverse('django_ledger:je-create', kwargs={
            'entity_slug': self.entity_slug,
            'ledger_pk': self.ledger_id
        })

    def get_detail_txs_url(self) -> str:
        """
        Generates the URL to view transaction details of the journal entry.

        Returns
        -------
        str
            The URL for transaction details of the journal entry.
        """
        return reverse('django_ledger:je-detail-txs', kwargs={
            'entity_slug': self.entity_slug,
            'ledger_pk': self.ledger_id,
            'je_pk': self.uuid
        })

    def get_unlock_url(self) -> str:
        """
        Generates the URL to mark the journal entry as unlocked.

        Returns
        -------
        str
            The URL for unlocking the journal entry.
        """
        return reverse('django_ledger:je-mark-as-unlocked', kwargs={
            'entity_slug': self.entity_slug,
            'ledger_pk': self.ledger_id,
            'je_pk': self.uuid
        })

    def get_lock_url(self) -> str:
        """
        Generates the URL to mark the journal entry as locked.

        Returns
        -------
        str
            The URL for locking the journal entry.
        """
        return reverse('django_ledger:je-mark-as-locked', kwargs={
            'entity_slug': self.entity_slug,
            'ledger_pk': self.ledger_id,
            'je_pk': self.uuid
        })

    def get_post_url(self) -> str:
        """
        Generates the URL to mark the journal entry as posted.

        Returns
        -------
        str
            The URL for posting the journal entry.
        """
        return reverse('django_ledger:je-mark-as-posted', kwargs={
            'entity_slug': self.entity_slug,
            'ledger_pk': self.ledger_id,
            'je_pk': self.uuid
        })

    def get_unpost_url(self) -> str:
        """
        Generates the URL to mark the journal entry as unposted.

        Returns
        -------
        str
            The URL for unposting the journal entry.
        """
        return reverse('django_ledger:je-mark-as-unposted', kwargs={
            'entity_slug': self.entity_slug,
            'ledger_pk': self.ledger_id,
            'je_pk': self.uuid
        })

    # Action URLS....
    def get_action_post_url(self) -> str:
        """
        Generates the URL used to mark the journal entry as posted.

        Returns
        -------
        str
            The generated URL for marking the journal entry as posted.
        """
        return reverse('django_ledger:je-mark-as-posted',
                       kwargs={
                           'entity_slug': self.entity_slug,
                           'ledger_pk': self.ledger_id,
                           'je_pk': self.uuid
                       })

    def get_action_unpost_url(self) -> str:
        """
        Generates the URL used to mark the journal entry as unposted.

        Returns
        -------
        str
            The generated URL for marking the journal entry as unposted.
        """
        return reverse('django_ledger:je-mark-as-unposted',
                       kwargs={
                           'entity_slug': self.entity_slug,
                           'ledger_pk': self.ledger_id,
                           'je_pk': self.uuid
                       })

    def get_action_lock_url(self) -> str:
        """
        Generates the URL used to mark the journal entry as locked.

        Returns
        -------
        str
            The generated URL for marking the journal entry as locked.
        """
        return reverse('django_ledger:je-mark-as-locked',
                       kwargs={
                           'entity_slug': self.entity_slug,
                           'ledger_pk': self.ledger_id,
                           'je_pk': self.uuid
                       })

    def get_action_unlock_url(self) -> str:
        """
        Generates the URL used to mark the journal entry as unlocked.

        Returns
        -------
        str
            The generated URL for marking the journal entry as unlocked.
        """
        return reverse('django_ledger:je-mark-as-unlocked',
                       kwargs={
                           'entity_slug': self.entity_slug,
                           'ledger_pk': self.ledger_id,
                           'je_pk': self.uuid
                       })


class JournalEntryModel(JournalEntryModelAbstract):
    """
    Journal Entry Model Base Class From Abstract
    """

    class Meta(JournalEntryModelAbstract.Meta):
        abstract = False


def journalentrymodel_presave(instance: JournalEntryModel, **kwargs):
    if instance._state.adding:
        # cannot add journal entries to a locked ledger...
        if instance.ledger_is_locked():
            raise JournalEntryValidationError(
                message=_(f'Cannot add Journal Entries to locked LedgerModel {instance.ledger_id}')
            )
    instance.generate_je_number(commit=False)


pre_save.connect(journalentrymodel_presave, sender=JournalEntryModel)



import base64
import hashlib
import json
import random
import string
from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Creates a code challenge and a code verifier used for Authorization Code grants when using OAuth2'

    def add_arguments(self, parser):
        parser.add_argument('--output', type=str, choices=['stdout', 'json'], default='stdout')

    def handle(self, *args, **options):
        code_verifier = ''.join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(random.randint(43, 128)))
        code_verifier = base64.urlsafe_b64encode(code_verifier.encode('utf-8'))

        code_challenge = hashlib.sha256(code_verifier).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge).decode('utf-8').replace('=', '')

        code_verifier = code_verifier.decode("utf-8")

        if options['output'] == 'stdout':
            self.stdout.write(self.style.SUCCESS(f'Code Verifier: {code_verifier}'))
            self.stdout.write(self.style.SUCCESS(f'Code Challenge: {code_challenge}'))

        elif options['output'] == 'json':
            out_file = Path(settings.BASE_DIR).joinpath('oauth_codes.json')

            codes = {
                'code_verifier': code_verifier,
                'code_challenge': code_challenge
            }

            with open(out_file, 'w') as io:
                json.dump(codes, io, indent=4)



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/jobs/',
         views.ImportJobModelListView.as_view(),
         name='data-import-jobs-list'),
    path('<slug:entity_slug>/jobs/<uuid:job_pk>/update/',
         views.ImportJobModelUpdateView.as_view(),
         name='data-import-jobs-update'),
    path('<slug:entity_slug>/jobs/<uuid:job_pk>/delete/',
         views.ImportJobModelDeleteView.as_view(),
         name='data-import-jobs-delete'),
    path('<slug:entity_slug>/import-ofx/',
         views.ImportJobModelCreateView.as_view(),
         name='data-import-ofx'),
    path('<slug:entity_slug>/jobs/<uuid:job_pk>/txs/',
         views.DataImportJobDetailView.as_view(),
         name='data-import-job-txs'),
]
